# -*- coding: utf-8 -*-
"""Test for the base Dimension class."""
import numpy as np
import pytest
from mrsimulator import Dimension
from pydantic import ValidationError


def test_direct_init_dimension():
    the_dimension = Dimension(number_of_points=1024, spectral_width=100)

    assert the_dimension.number_of_points == 1024
    error = "ensure this value is greater than 0"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        the_dimension.number_of_points = -1024

    # spectral width test
    assert the_dimension.spectral_width == 100
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        the_dimension.spectral_width = 0
    # ensure the default value is Hz
    assert the_dimension.property_units["spectral_width"] == "Hz"

    assert the_dimension.spin is None
    assert the_dimension.natural_abundance is None
    assert the_dimension.gyromagnetic_ratio is None
    assert the_dimension.quadrupole_moment is None
    assert the_dimension.atomic_number is None
    assert the_dimension.larmor_frequency is None

    # rotor angle test
    assert the_dimension.rotor_angle == 0.9553166
    # ensure the unit of rotor angle is radian
    assert the_dimension.property_units["rotor_angle"] == "rad"
    the_dimension.rotor_angle = 1.5707963268
    assert the_dimension.rotor_angle == 1.5707963268
    # rotor angle cannot ve greater than pi/2 and less that 0
    error = "ensure this value is less than or equal to 1.5707963268"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        the_dimension.rotor_angle = 2.57079
    error = "ensure this value is greater than or equal to 0"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        the_dimension.rotor_angle = -0.1243

    # rotor frequency test
    assert the_dimension.rotor_frequency == 0
    the_dimension.rotor_frequency = 1000
    assert the_dimension.rotor_frequency == 1000
    assert the_dimension.property_units["rotor_frequency"] == "Hz"

    # magnetic flux density
    assert the_dimension.magnetic_flux_density == 9.4
    assert the_dimension.property_units["magnetic_flux_density"] == "T"
    error = "ensure this value is greater than or equal to 0"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        the_dimension.magnetic_flux_density = -1

    # isotope
    assert the_dimension.isotope is None
    the_dimension.isotope = "13C"
    print(the_dimension)
    assert the_dimension.isotope == "13C"
    assert the_dimension.gyromagnetic_ratio == 10.7084
    assert the_dimension.spin == 0.5
    assert the_dimension.larmor_frequency == -10.7084 * 9.4 * 1e6

    coordinate_Hz = (np.arange(1024) - 512) * (100.0 / 1024)
    assert np.allclose(the_dimension.coordinates_Hz, coordinate_Hz)

    coordinate_ppm = coordinate_Hz / (10.7084 * 9.4)
    assert np.allclose(the_dimension.coordinates_ppm, coordinate_ppm)

    error = "field required"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        Dimension()

    # when isotope is None
    the_dimension.isotope = None
    assert the_dimension.isotope is None
    assert the_dimension.spin is None
    assert the_dimension.natural_abundance is None
    assert the_dimension.gyromagnetic_ratio is None
    assert the_dimension.quadrupole_moment is None
    assert the_dimension.atomic_number is None
    assert the_dimension.larmor_frequency is None

    assert np.allclose(the_dimension.coordinates_Hz, coordinate_Hz)
    assert the_dimension.coordinates_ppm is None

    error = "Could not parse isotope string"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        Dimension(isotope="monkey", spectral_width=10000)

    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        Dimension(isotope="15H", spectral_width=10000)

    result = {
        "number_of_points": 1024,
        "spectral_width": "100.0 Hz",
        "reference_offset": "0 Hz",
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "1000.0 Hz",
        "rotor_angle": "1.5707963268 rad",
        "label": "",
    }
    assert the_dimension.to_dict_with_units() == result

    the_dimension.isotope = "27Al"
    result = {
        "isotope": "27Al",
        "number_of_points": 1024,
        "spectral_width": "100.0 Hz",
        "reference_offset": "0 Hz",
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "1000.0 Hz",
        "rotor_angle": "1.5707963268 rad",
        "label": "",
    }
    assert the_dimension.to_dict_with_units() == result


def test_parse_json_spectrum():
    good_json = {
        "number_of_points": 1024,
        "spectral_width": "100 Hz",
        "reference_offset": "0 Hz",
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "0 Hz",
        "rotor_angle": "54.935 degree",  # 54.935 degrees in radians
        "isotope": "1H",
    }

    spec = Dimension.parse_dict_with_units(good_json)
    assert spec.spin == 0.5
    assert spec.isotope == "1H"
    assert spec.gyromagnetic_ratio == 42.57748
    assert spec.natural_abundance == 99.985
    assert spec.quadrupole_moment == 0.0
    assert spec.atomic_number == 1
    assert np.allclose(spec.rotor_angle, 0.95879662)


def test_parse_json_spectrum2():
    good_json = {
        "number_of_points": 1024,
        "spectral_width": "100 Hz",
        "reference_offset": "0 Hz",
        "magnetic_flux_density": "9.4 T",
        "rotor_frequency": "0 Hz",
        "rotor_angle": "54.935 degree",  # 54.935 degrees in radians
        "rotor_phase": "0 rad",
        "isotope": "29Si",
    }

    spec = Dimension.parse_dict_with_units(good_json)
    assert spec.spin == 0.5
    assert spec.isotope == "29Si"
    assert spec.gyromagnetic_ratio == -8.465499
    assert spec.natural_abundance == 4.683
    assert spec.quadrupole_moment == 0.0
    assert spec.atomic_number == 14
    assert np.allclose(spec.rotor_angle, 0.95879662)



# -*- coding: utf-8 -*-
#
#  test.pxd
#
#  Created by Deepansh J. Srivastava.
#  Contact email = deepansh2012@gmail.com
#

from libcpp cimport bool as bool_t

cdef extern from "angular_momentum.h":
    void wigner_d_matrices(const int l, const int n, const double *angle, double *wigner)

    void wigner_d_matrices_from_exp_I_beta(const int l, const int n, const double complex *exp_I_beta,
                                  double *wigner)

    # void __wigner_rotation(const int l, const int n, const double *wigner, const double *cos_alpha,
    #                        const double complex *R_in, double complex *R_out)

    void __wigner_rotation_2(const int l, const int n, const double *wigner,
                             const void *exp_Im_alpha, const void *R_in,
                             void *R_out)

    void single_wigner_rotation(const int l, const double *euler_angles,
                            const void *R_in, void *R_out)

    void wigner_dm0_vector(const int l, const double beta, double *R_out)

    void get_exp_Im_alpha(const unsigned int octant_orientations,
                          const bool_t allow_fourth_rank, void *exp_Im_alpha)

    void __batch_wigner_rotation(const unsigned int octant_orientations,
                             const unsigned int n_octants,
                             double *wigner_2j_matrices,
                             void *R2,
                             double *wigner_4j_matrices,
                             void *R4,
                             void *exp_Im_alpha,
                             void *w2, void *w4)


cdef extern from "powder_setup.h":
    void octahedron_averaging_setup(
        int nt,
        double complex *exp_I_alpha,
        double complex *exp_I_beta,
        double *amp)

cdef extern from "interpolation.h":
    void triangle_interpolation(
        double *freq1,
        double *freq2,
        double *freq3,
        double *amp,
        double *spec,
        int *points)

cdef extern from "octahedron.h":
    void octahedronInterpolation(
        double *spec,
        double *freq,
        int nt,
        double *amp,
        int stride,
        int m)

cdef extern from "mrsimulator.h":
    void __get_components(
        int number_of_sidebands,
        double spin_frequency,
        double *pre_phase)

#     ctypedef struct MRS_plan

#     MRS_plan *MRS_create_plan(
#         unsigned int geodesic_polyhedron_frequency,
#         int number_of_sidebands,
#         double sample_rotation_frequency_in_Hz,
#         double rotor_angle_in_rad, double increment,
#         bool_t allow_fourth_rank)


cdef extern from "isotopomer_ravel.h":
    ctypedef struct isotopomer_ravel:
        int number_of_sites;                    # Number of sites
        float spin;                             # The spin quantum number
        double larmor_frequency;                # Larmor frequency (MHz)
        double *isotropic_chemical_shift_in_Hz; # Isotropic chemical shift (Hz)
        double *shielding_anisotropy_in_Hz;     # Nuclear shielding anisotropy (Hz)
        double *shielding_asymmetry;            # Nuclear shielding asymmetry parameter
        double *shielding_orientation;          # Nuclear shielding PAS to CRS euler angles (rad.)
        double *quadrupole_coupling_constant_in_Hz;     # Quadrupolar coupling constant (Hz)
        double *quadrupole_asymmetry;          # Quadrupolar asymmetry parameter
        double *quadrupole_orientation;        # Quadrupolar PAS to CRS euler angles (rad.)
        double *dipolar_couplings;              # dipolar coupling stored as list of lists

    ctypedef struct isotopomers_list:
        isotopomer_ravel *isotopomers

cdef extern from "spinning_sidebands.h":
    void spinning_sideband_core(
        # spectrum information and related amplitude
        double * spec,
        double spectral_start,
        double spectral_increment,
        int number_of_points,

        isotopomer_ravel *ravel_isotopomer,

        int quad_second_order,                    # Quad theory for second order,
        int remove_second_order_quad_isotropic,   # remove the isotropic contribution from the
                                                  # second order quad Hamiltonian.

        # spin rate, spin angle and number spinning sidebands
        int number_of_sidebands,
        double sample_rotation_frequency_in_Hz,
        double rotor_angle_in_rad,

        # The transition as transition[0] = mi and transition[1] = mf
        double *transition,
        int geodesic_polyhedron_frequency,
        unsigned int integration_volume             # 0-octant, 1-hemisphere, 2-sphere.
        )



# -*- coding: utf-8 -*-
"""Base Parseable class."""
from typing import ClassVar
from typing import Dict

from csdmpy.units import string_to_quantity
from pydantic import BaseModel

__author__ = "Shyam Dwaraknath"
__email__ = "shyamd@lbl.gov"


class Parseable(BaseModel):
    """
    Base class for all objects that can be parsed easily from JSON with units
    Don't directly use this. Rather inherit from it and implement a data model
    and property units and defaults
    """

    property_unit_types: ClassVar = {}

    property_default_units: ClassVar = {}

    property_units: Dict = {}

    @classmethod
    def parse_dict_with_units(cls, json_dict):
        # Only consider properties with both unit types and default units
        all_props = set(cls.property_unit_types.keys()).intersection(
            set(cls.property_default_units.keys())
        )

        property_units = {}
        for prop in all_props:
            required_type = cls.property_unit_types[prop]
            default_unit = cls.property_default_units[prop]

            if prop in json_dict:
                # If we have a single option
                if isinstance(required_type, str) and isinstance(default_unit, str):
                    try:
                        json_dict[prop] = enforce_units(
                            json_dict[prop], required_type, default_unit
                        )
                        property_units[prop] = default_unit
                    except Exception as e:
                        raise Exception(
                            f"Error enforcing units for {prop}: {json_dict[prop]}\n"
                            + str(e)
                        )

                # If there are multiple type/unit combinations
                elif isinstance(required_type, list) and isinstance(default_unit, list):
                    pos_values = [
                        enforce_units(
                            json_dict[prop], r_type, d_unit, throw_error=False
                        )
                        for r_type, d_unit in zip(required_type, default_unit)
                    ]
                    # If none of the units were enforceable, error
                    # else choose the first good one
                    if not ([val is not None for val in pos_values]):
                        raise Exception(f"Could not enforce any units on {prop}")
                    else:
                        json_dict[prop], property_units[prop] = [
                            d for d in zip(pos_values, default_unit) if d[0] is not None
                        ][0]
        return cls(**json_dict, property_units=property_units)


def enforce_units(value: str, required_type: str, default_unit: str, throw_error=True):
    """
    Enforces a required type and default unit on the value

        value
    """
    try:
        value = string_to_quantity(value)
        data_type = value.unit.physical_type

        if required_type != data_type:
            raise Exception(
                f"A {required_type} value is required but got a {data_type} instead"
            )

        return value.to(default_unit).value
    except Exception as e:
        if throw_error:
            raise e
        else:
            return None



# -*- coding: utf-8 -*-
"""Base Simulator class."""
from typing import List
from typing import Optional

import csdmpy as cp
from astropy import units as u
from mrsimulator import Dimension
from mrsimulator import Isotopomer
from mrsimulator.importer import import_json
from mrsimulator.simulator_config import ConfigSimulator
from pydantic import BaseModel

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class Simulator(BaseModel):
    """
    The simulator class.

    Attributes:
        isotopomers: List of :ref:`isotopomer_api` objects.
        dimensions: List of :ref:`dimension_api` objects.
        config: :ref:`config_api` object.
    """

    isotopomers: List[Isotopomer] = []
    dimensions: List[Dimension] = []
    simulated_data: Optional[List]
    config: ConfigSimulator = ConfigSimulator()

    class Config:
        validate_assignment = True
        arbitrary_types_allowed = True

    def __eq__(self, other):
        if (
            not isinstance(other, Simulator)
            or self.isotopomers != other.isotopomers
            or self.dimensions != other.dimensions
            or self.config != other.config
        ):
            return False
        return True

    def get_isotopes(self, I=None):
        """
        Set of unique isotopes from sites in the list of isotopomers corresponding
        to spin quantum number `I`. If `I` is unspecified or None, a set of all
        unique isotopes is returned instead.

        Args:
            I: (optional) The spin quantum number. Valid input are multiples of 0.5.

        Returns:
            A Set

        Example:
            >>> sim.get_isotopes() # doctest:+SKIP
            {'1H', '27Al', '13C'}
            >>> sim.get_isotopes(I=0.5) # doctest:+SKIP
            {'1H', '13C'}
            >>> sim.get_isotopes(I=1.5)
            set()
            >>> sim.get_isotopes(I=2.5)
            {'27Al'}
        """
        st = set()
        for isotopomer in self.isotopomers:
            st.update(isotopomer.get_isotopes(I))
        return st

    def to_dict_with_units(self, include_dimensions=False):
        """
        Serialize the isotopomers and dimensions attribute from the Simulator object
        to a JSON compliant python dictionary with units.

        Args:
            remove_dimensions: A boolean. If True, the output contains serialized
                dimension objects. Default is False.

        Returns:
            Dict object

        Example:
            >>> pprint(sim.to_dict_with_units())
            {'isotopomers': [{'abundance': '100%',
                              'description': '',
                              'name': '',
                              'sites': [{'isotope': '13C',
                                         'isotropic_chemical_shift': '20.0 ppm',
                                         'shielding_symmetric': {'eta': 0.5,
                                                                 'zeta': '10.0 ppm'}}]},
                             {'abundance': '100%',
                              'description': '',
                              'name': '',
                              'sites': [{'isotope': '1H',
                                         'isotropic_chemical_shift': '-4.0 ppm',
                                         'shielding_symmetric': {'eta': 0.1,
                                                                 'zeta': '2.1 ppm'}}]},
                             {'abundance': '100%',
                              'description': '',
                              'name': '',
                              'sites': [{'isotope': '27Al',
                                         'isotropic_chemical_shift': '120.0 ppm',
                                         'shielding_symmetric': {'eta': 0.1,
                                                                 'zeta': '2.1 ppm'}}]}]}
        """
        sim = {}
        sim["isotopomers"] = [_.to_dict_with_units() for _ in self.isotopomers]

        if include_dimensions:
            dimensions = [_.to_dict_with_units() for _ in self.dimensions]
            if dimensions != []:
                sim["dimensions"] = dimensions

        return sim

    def load_isotopomers(self, filename):
        """
        Load a list of isotopomers from JSON serialized isotopomers file.

        See an
        `example <https://raw.githubusercontent.com/DeepanshS/mrsimulator-test
        /master/isotopomers_ppm.json>`_
        of JSON serialized isotopomers file. For details, refer to the
        :ref:`load_isotopomers` section.

        Args:
            `filename`: A local or remote address to the JSON serialized isotopomers
                        file.
        Example:
            >>> sim.load_isotopomers(filename) # doctest:+SKIP
        """
        contents = import_json(filename)
        json_data = contents["isotopomers"]
        self.isotopomers = [Isotopomer.parse_dict_with_units(obj) for obj in json_data]

    def run(self, method, **kwargs):
        """Simulate the lineshape.

        Args:
            method: The methods used in simulating line-shapes.

        Example:
            >>> sim.run(method=one_d_spectrum) # doctest:+SKIP
        """
        isotopomers = [
            isotopomer.to_freq_dict(self.dimensions[0].magnetic_flux_density)
            for isotopomer in self.isotopomers
        ]

        amp = method(
            dimension=self.dimensions,
            isotopomers=isotopomers,
            **self.config._dict,
            **kwargs,
        )

        if isinstance(amp, list):
            self.simulated_data = amp
        else:
            self.simulated_data = [amp]

        """The frequency is in the units of Hz."""
        freq = self.dimensions[0].coordinates_ppm
        freq *= u.Unit("ppm")
        return freq, amp

    def as_csdm_object(self):
        """
        Converts the data to a CSDM object. Read
        `csdmpy <https://csdmpy.readthedocs.io/en/latest/>`_ for details

        Return:
            CSDM object
        """
        new = cp.new()
        for dimension in self.dimensions:
            new_dimension = {
                "type": "linear",
                "count": dimension.number_of_points,
                "increment": "{0} Hz".format(
                    dimension.spectral_width / dimension.number_of_points
                ),
                "coordinates_offset": f"{dimension.reference_offset} Hz",
                "origin_offset": f"{dimension.larmor_frequency} Hz",
                "complex_fft": True,
            }
            new.add_dimension(new_dimension)

        dependent_variable = {
            "type": "internal",
            "quantity_type": "scalar",
            "numeric_type": "float64",
        }
        for index, datum in enumerate(self.simulated_data):
            if datum != []:
                dependent_variable["components"] = [datum]
                name = self.isotopomers[index].name
                if name != "":
                    dependent_variable.update({"name": name})

                description = self.isotopomers[index].description
                if description != "":
                    dependent_variable.update({"description": description})

                dependent_variable["application"] = {
                    "com.github.DeepanshS.mrsimulator": {
                        "isotopomers": [self.isotopomers[index].to_dict_with_units()]
                    }
                }
                new.add_dependent_variable(dependent_variable)
                new.dependent_variables[-1].encoding = "base64"
        return new



# -*- coding: utf-8 -*-
"""Test for the base Simulator class."""
from random import randint

import numpy as np
import pytest
from mrsimulator import Dimension
from mrsimulator import Isotopomer
from mrsimulator import Simulator
from mrsimulator import Site
from mrsimulator.methods import one_d_spectrum


def test_simulator_assignments():
    a = Simulator()
    assert a.isotopomers == []

    error = "value is not a valid list"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        a.isotopomers = ""

    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        a.dimensions = ""


def test_equality():
    a = Simulator()
    b = Simulator()
    assert a == b

    assert a is not {}

    c = Simulator(isotopomers=[Isotopomer()])
    assert a is not c

    result = {
        "isotopomers": [
            {"abundance": "100%", "description": "", "name": "", "sites": []}
        ]
    }
    assert c.to_dict_with_units(include_dimensions=True) == result

    result["dimensions"] = [
        {
            "label": "",
            "magnetic_flux_density": "9.4 T",
            "number_of_points": 1024,
            "reference_offset": "0 Hz",
            "rotor_angle": "0.9553166 rad",
            "rotor_frequency": "0 Hz",
            "spectral_width": "10.0 Hz",
        }
    ]
    c.dimensions = [Dimension(spectral_width=10)]
    assert c.to_dict_with_units(include_dimensions=True) == result


def get_simulator():
    isotopes = ["19F", "31P", "2H", "6Li", "14N", "27Al", "25Mg", "45Sc", "87Sr"]
    sites = []
    for isotope in isotopes:
        for _ in range(randint(1, 3)):
            sites.append(Site(isotope=isotope))
    sim = Simulator()
    sim.isotopomers.append(Isotopomer(sites=sites))
    return sim


def test_get_isotopes():
    isotopes = {"19F", "31P", "2H", "6Li", "14N", "27Al", "25Mg", "45Sc", "87Sr"}
    sim = get_simulator()
    assert sim.get_isotopes() == isotopes
    assert sim.get_isotopes(I=0.5) == {"19F", "31P"}
    assert sim.get_isotopes(I=1) == {"2H", "6Li", "14N"}
    assert sim.get_isotopes(I=1.5) == set()
    assert sim.get_isotopes(I=2.5) == {"27Al", "25Mg"}
    assert sim.get_isotopes(I=3.5) == {"45Sc"}
    assert sim.get_isotopes(I=4.5) == {"87Sr"}


def test_csdm_object():
    sim = Simulator()
    site = Site(
        isotope="27Al",
        isotropic_chemical_shift=120,
        shielding_symmetric={"zeta": 2.1, "eta": 0.1},
        quadrupole={"Cq": 5.1e6, "eta": 0.5},
    )
    sim.isotopomers = [{"name": "test", "description": "awesome", "sites": [site]}]
    sim.dimensions = [
        {
            "number_of_points": 1024,
            "spectral_width": 100,
            "reference_offset": 0,
            "magnetic_flux_density": 9.4,
            "rotor_frequency": 0,
            "rotor_angle": 0.9553166,
            "isotope": "27Al",
        }
    ]
    x, y = sim.run(method=one_d_spectrum)
    csdm_obj = sim.as_csdm_object()

    assert np.allclose(csdm_obj.dependent_variables[0].components[0], y)
    assert csdm_obj.dimensions[0].count == x.size



# -*- coding: utf-8 -*-
"""Base Site class."""
from typing import ClassVar
from typing import Dict
from typing import Optional

from mrsimulator import AntisymmetricTensor
from mrsimulator import Parseable
from mrsimulator import SymmetricTensor
from mrsimulator.dimension import get_isotope_data
from pydantic import validator

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class Site(Parseable):
    """
    Base Site class representing a nuclear isotope.

    Arguments:
        name: An optional string with a name or id of the site.
        isotope: An optional string expressed as atomic number followed by an
                isotope symbol, eg. ``13C``, ``17O``. The default value is ``1H``.
        isotropic_chemical_shift: An optional floating point number representing
                the isotropic chemical shift of the site in unit of ppm. The
                default value is ``0``.
        shielding_symmetric: An optional SymmetricTensor object or an equivalent
                python dict object representing the irreducible second-rank traceless
                symmetric part of the nuclear shielding tensor. The default value is
                ``None``.
        shielding_antisymmetric: An optional AntisymmetricTensor object or an
                equivalent python dict object representing the irreducible first-rank
                antisymmetric part of the nuclear shielding tensor. The default value
                is ``None``.
        quadrupolar: An optional SymmetricTensor object or an equivalent python dict
                object representing the irreducible second-rank traceless symmetric
                part of electric-field gradient tensor. The default value is ``None``.

    Example:
        Setting up Site objects.

            >>> site1 = Site(
            ...           isotope='13C',
            ...           isotropic_chemical_shift=20, # in ppm
            ...           shielding_symmetric={
            ...             "zeta": 10, # in ppm
            ...             "eta": 0.5
            ...           },
            ...           quadrupole={
            ...             "Cq": 5.1e6, # in Hz
            ...             "eta": 0.5
            ...           }
            ...         )

        Using SymmetricTensor objects.

        .. doctest::

            >>> site1 = Site(
            ...           isotope='13C',
            ...           isotropic_chemical_shift=20, # in ppm
            ...           shielding_symmetric=SymmetricTensor(zeta=10, eta=0.5),
            ...           quadrupole=SymmetricTensor(Cq=5.1e6, eta=0.5)
            ...         )
    """

    name: str = None
    isotope: str = "1H"
    isotropic_chemical_shift: Optional[float] = 0
    shielding_symmetric: Optional[SymmetricTensor] = None
    shielding_antisymmetric: Optional[AntisymmetricTensor] = None
    quadrupolar: Optional[SymmetricTensor] = None

    property_unit_types: ClassVar = {"isotropic_chemical_shift": "dimensionless"}
    property_default_units: ClassVar = {"isotropic_chemical_shift": "ppm"}
    property_units: Dict = {"isotropic_chemical_shift": "ppm"}

    @validator("quadrupolar")
    def spin_must_be_at_least_one(cls, v, values):
        isotope = values["isotope"]
        I = get_isotope_data(isotope)["spin"] / 2.0
        if I >= 1:
            if "zeta" in v.property_units:
                v.property_units.pop("zeta")
            return v
        message = f"with spin quantum number {I} does not allow quadrupolar tensor."
        raise ValueError(f"{isotope} {message}")

    @validator("shielding_symmetric")
    def shielding_symmetric_must_not_contain_Cq(cls, v, values):
        if "Cq" in v.property_units:
            v.property_units.pop("Cq")
        return v

    class Config:
        validate_assignment = True

    @property
    def spin(self):
        """Spin quantum number of the isotope."""
        return get_isotope_data(self.isotope)["spin"] / 2.0

    @property
    def natural_abundance(self):
        """Natural abundance of the isotope in units of %."""
        return get_isotope_data(self.isotope)["natural_abundance"]

    @property
    def gyromagnetic_ratio(self):
        """Reduced gyromagnetic ratio of the isotope in units of MHz/T."""
        return get_isotope_data(self.isotope)["gyromagnetic_ratio"]

    @property
    def quadrupole_moment(self):
        """Quadrupole moment of the isotope in units of eB (electron-Barn)."""
        return get_isotope_data(self.isotope)["quadrupole_moment"]

    @property
    def atomic_number(self):
        """Atomic number of the isotope."""
        return get_isotope_data(self.isotope)["atomic_number"]

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantities from a dictionary representation of the Site
        object.

        Args:
            py_dict: A required Dict object.

        Returns:
            Site object

        Example:
            >>> site_dict = {
            ...    "isotope": "13C",
            ...    "isotropic_chemical_shift": "20 ppm",
            ...    "shielding_symmetric": {"zeta": "10 ppm", "eta":0.5}
            ... }
            >>> site1 = Site.parse_dict_with_units(site_dict)
        """
        prop_mapping = {
            "shielding_symmetric": SymmetricTensor,
            "shielding_antisymmetric": AntisymmetricTensor,
            "quadrupolar": SymmetricTensor,
        }

        for k, v in prop_mapping.items():
            if k in py_dict:
                py_dict[k] = v.parse_dict_with_units(py_dict[k])

        return super().parse_dict_with_units(py_dict)

    def to_freq_dict(self, B0):
        """
        Serialize the Site object to a JSON compliant python dictionary where the
        attribute values are numbers expressed in default units. The default unit
        for attributes with respective dimensionalities are:
        - frequency: `Hz`
        - angle: `rad`

        Args:
            B0: A required macroscopic magnetic flux density given in units of T.

        Return:
            Dict object

        Example:
            >>> pprint(site1.to_freq_dict(9.4))
            {'isotope': '13C',
             'isotropic_chemical_shift': -2013.1791999999998,
             'name': None,
             'quadrupolar': None,
             'shielding_antisymmetric': None,
             'shielding_symmetric': {'alpha': None,
                                     'beta': None,
                                     'eta': 0.5,
                                     'gamma': None,
                                     'zeta': -1006.5895999999999}}
        """
        temp_dict = self.dict()
        larmor_frequency = -self.gyromagnetic_ratio * B0  # in MHz
        for k in ["shielding_symmetric", "shielding_antisymmetric", "quadrupolar"]:
            if getattr(self, k):
                temp_dict[k] = getattr(self, k).to_freq_dict(larmor_frequency)
                if k == "shielding_symmetric":
                    temp_dict[k].pop("Cq")

        temp_dict["isotropic_chemical_shift"] *= larmor_frequency
        temp_dict.pop("property_units")
        return temp_dict

    def to_dict_with_units(self):
        """
        Serialize the Site object to a JSON compliant python dictionary with units.

        Returns:
            Dict object

        Example:
            >>> pprint(site1.to_dict_with_units())
            {'isotope': '13C',
             'isotropic_chemical_shift': '20.0 ppm',
             'shielding_symmetric': {'eta': 0.5, 'zeta': '10.0 ppm'}}

        """
        temp_dict = {k: v for k, v in self.dict().items() if v is not None}

        for k in ["shielding_symmetric", "shielding_antisymmetric", "quadrupolar"]:
            if getattr(self, k):
                temp_dict[k] = getattr(self, k).to_dict_with_units()

        temp_dict["isotropic_chemical_shift"] = (
            str(temp_dict["isotropic_chemical_shift"]) + " ppm"
        )
        temp_dict.pop("property_units")

        return temp_dict



# -*- coding: utf-8 -*-
import numpy as np

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


def pre_phase_components(number_of_sidebands, spin_frequency):
    r"""
    Calculates a 2D matrix `pre_phase` of shape (number_of_sidebands, 9) with
    the following definition

    `pre_phase = $2i\pi \frac{(\exp(i m \omega_r t) - 1)}{i m \omega_r}$`

    where m goes from -4 to 4 and t is the time points of size
    `number_of_points`. The time increment is given as
    `tau = 1.0 / (number_of_sidebands * spin_frequency)`.
    """
    n = int(number_of_sidebands)
    m_wr = spin_frequency * 2.0 * np.pi * (np.arange(9) - 4.0)
    time = np.arange(n) / (number_of_sidebands * spin_frequency)
    time, m_wr = np.meshgrid(time, m_wr)
    pht = 1j * time * m_wr
    pre_phase = np.empty(pht.shape, np.complex128)
    np.expm1(pht, out=pre_phase)
    pre_phase[0:4] *= (2.0 * np.pi) / m_wr[0:4]
    pre_phase[5:9] *= (2.0 * np.pi) / m_wr[5:9]
    pre_phase[4] = 0.0
    return pre_phase



# -*- coding: utf-8 -*-
"""Base Isotopomer class."""
from copy import deepcopy
from typing import ClassVar
from typing import Dict
from typing import List
from typing import Optional

from mrsimulator import Parseable
from mrsimulator.dimension import ISOTOPE_DATA
from mrsimulator.site import Site
from pydantic import Field

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class Isotopomer(Parseable):
    """
    Base isotopmer class representing an isolated spin-system with sites and couplings.

    Arguments:
        name: An optional string with the isotopomer name. The default is an empty
                string.
        description: An optional string describing the isotopomer. The default is
                an empty string.
        sites: A list of :ref:`site_api` objects or an equivalent python dict object
                representing a nuclear site. Default value is an empty list.
        abundance: The abundance of the isotopomer in unit of %. The default value is
                100. This attribute is useful when multiple isotopomers are present.
    """

    name: Optional[str] = ""
    description: Optional[str] = ""
    sites: List[Site] = []
    # couplings: list = [], # TODO: Deepansh what should this look like?
    abundance: float = Field(default=100, ge=0, le=100)

    property_unit_types: ClassVar = {"abundance": "dimensionless"}
    property_default_units: ClassVar = {"abundance": "pct"}
    property_units: Dict = {"abundance": "pct"}

    class Config:
        validate_assignment = True

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse physical quantity from the attributes of an isotopomer object
        expressed as a python dictionary. The physical quantities are expressed
        as string with a number followed by a unit.

        Args:
            py_dict: Python dictionary representation of an isotopomer object with
                    attributes values as string with a physical quantity.

        Example:
            >>> isotopomer_dict = {
            ...     "sites": [{
            ...         "isotope":"13C",
            ...         "isotropic_chemical_shift": "20 ppm",
            ...         "shielding_symmetric": {
            ...             "zeta": "10 ppm",
            ...             "eta": 0.5
            ...         }
            ...     }]
            ... }
            >>> isotopomer_1 = Isotopomer.parse_dict_with_units(isotopomer_dict)
        """
        py_dict_copy = deepcopy(py_dict)
        if "sites" in py_dict_copy:
            py_dict_copy["sites"] = [
                Site.parse_dict_with_units(s) for s in py_dict_copy["sites"]
            ]

        return super().parse_dict_with_units(py_dict_copy)

    def to_freq_dict(self, B0):
        """
        Serialize the Isotopomer object to a JSON compliant python dictionary where the
        attribute values are numbers expressed in default units. The default unit
        for attributes with respective dimensionalities are:
        - frequency: `Hz`
        - angle: `rad`

        Args:
            B0: The macroscopic magnetic flux density in units of Tesla, T.

        Return: A python dict

        Example:
            >>> pprint(isotopomer_1.to_freq_dict(B0=9.4))
            {'abundance': 100,
             'description': '',
             'name': '',
             'sites': [{'isotope': '13C',
                        'isotropic_chemical_shift': -2013.1791999999998,
                        'name': None,
                        'quadrupolar': None,
                        'shielding_antisymmetric': None,
                        'shielding_symmetric': {'alpha': None,
                                                'beta': None,
                                                'eta': 0.5,
                                                'gamma': None,
                                                'zeta': -1006.5895999999999}}]}
        """
        temp_dict = self.dict()
        temp_dict["sites"] = [site.to_freq_dict(B0) for site in self.sites]
        temp_dict.pop("property_units")
        return temp_dict

    def to_dict_with_units(self):
        """
        Serialize the Isotopomer object to a JSON compliant python dictionary object
        where attribute values are physical quantities expressed as a string with a
        number followed by a unit.

        Return: A python dict

        Example:
            >>> pprint(isotopomer_1.to_dict_with_units())
            {'abundance': '100%',
             'description': '',
             'name': '',
             'sites': [{'isotope': '13C',
                        'isotropic_chemical_shift': '20.0 ppm',
                        'shielding_symmetric': {'eta': 0.5, 'zeta': '10.0 ppm'}}]}
        """
        temp_dict = self.dict()
        temp_dict["sites"] = [site.to_dict_with_units() for site in self.sites]
        temp_dict["abundance"] = f"{self.abundance}%"
        temp_dict.pop("property_units")

        return temp_dict

    def get_isotopes(self, I=None):
        """
        Set of unique isotopes from the list of sites corresponding to the given value
        of `I`. If `I` is unspecified or None, a set of all defined isotopes is
        returned instead.

        Args:
            I: (optional) The spin quantum number. Valid input are multiples of 0.5.

        Returns:
            A set

        Example:
            >>> isotopomers.get_isotopes() # doctest:+SKIP
            {'1H', '27Al', '13C'}
            >>> isotopomers.get_isotopes(I=0.5) # doctest:+SKIP
            {'1H', '13C'}
            >>> isotopomers.get_isotopes(I=1.5)
            set()
            >>> isotopomers.get_isotopes(I=2.5)
            {'27Al'}
        """
        return set(
            site.isotope for site in self.sites if site.isotope in allowed_isotopes(I)
        )


def allowed_isotopes(I=None):
    """
    List of NMR active isotopes currently supported in ``mrsimulator``.

    Args:
        I: (optional) The spin quantum number. Valid input are multiples of 0.5.

    Returns:
        A list of all isotopes supported in ``mrsimulator`` with the given spin
        quantum number `I`. If the spin is unspecified or None, a list of all
        allowed isotopes is returned instead.
    Returns:
        Set
    """
    if I is None:
        return list({isotope for isotope, data in ISOTOPE_DATA.items()})
    return list(
        {
            isotope
            for isotope, data in ISOTOPE_DATA.items()
            if data["spin"] == int(2 * I)
        }
    )



# -*- coding: utf-8 -*-
"""Base ConfigSimulator class."""
from copy import deepcopy

from mrsimulator.sandbox import AveragingScheme


__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class ConfigSimulator:
    """
    The configurable parametes used in lineshape simulation

    Attributes:
        number_of_sidebands: An integer with the number of sidebands requested in the
            lineshape simulation. This number cannot be zero or negative. The default
            value is ``64``.
        integration_volume: An enumeration literal with valid literals, 'octant',
            'hemisphere', over which the frequency contributions are evaluated. The
            default value if ``octant``.
        integration_density: An positive interger. If `n` is the integration_density,
            then the total number of orientation is given as
            ``(n+1)*(n+2)/2 * number of octant``. The line-shape is an integral over
            the frequency contribution arising from every orientation. The default
            value is ``70``.
        decompose: A boolean. If true, decomposes the line-shape into an array of
            line-shapes arising from individual isotopomer. If False, the lins-shape
            is a sum of individual line-shapes instead. The default value is ``False``.

    Example:
        >>> a = Simulator()
        >>> a.config.number_of_sidebands = 128
        >>> a.config.integration_volume = 'hemisphere'
        >>> a.config.decompose = True
    """

    integration_volume_enum = {"octant": 0, "hemisphere": 1}
    integration_volume_enum_rev = {0: "octant", 1: "hemisphere"}

    def __init__(self):
        self._dict = {
            "number_of_sidebands": 64,
            "integration_volume": 0,
            "integration_density": 70,
            "decompose": False,
        }

    def update_integration_scheme(self):
        self._averaging_scheme = AveragingScheme(
            integration_density=self._dict["integration_density"],
            integration_volume=self._dict["integration_volume"],
            allow_fourth_rank=True,
        )

    # decompose line-shape into array of lineshapes arising from individual
    # isotopomers.
    @property
    def decompose(self):
        return self._dict["decompose"]

    @decompose.setter
    def decompose(self, value):
        if isinstance(value, bool):
            self._dict["decompose"] = value
            return
        raise ValueError(f"Expecting a boolean.")

    # number of sidebands
    @property
    def number_of_sidebands(self):
        return self._dict["number_of_sidebands"]

    @number_of_sidebands.setter
    def number_of_sidebands(self, value):
        if isinstance(value, int):
            if value >= 0:
                self._dict["number_of_sidebands"] = value
                return
        raise ValueError(f"Expecting a positive integer, found {value}.")

    # integration density
    @property
    def integration_density(self):
        return self._dict["integration_density"]

    @integration_density.setter
    def integration_density(self, value):
        if isinstance(value, int):
            if value > 0:
                self._dict["integration_density"] = value
                self.update_integration_scheme()
                return
        raise ValueError(f"Expecting a positive integer, found {value}.")

    # integration volume
    @property
    def integration_volume(self):
        return self.__class__.integration_volume_enum_rev[
            self._dict["integration_volume"]
        ]

    @integration_volume.setter
    def integration_volume(self, value):
        class_dict = self.__class__.integration_volume_enum
        if value in class_dict:
            self._dict["integration_volume"] = class_dict[value]
            return
        raise ValueError(
            (
                "value is not a valid enumeration literal; "
                "permitted: 'octant', 'hemisphere', found {value}.",
            )
        )

    # averaging scheme. This contains the c pointer used in line-shape evaluation
    # at c-level.
    @property
    def averaging_scheme(self):
        return self._averaging_scheme

    # methods
    def __eq__(self, other):
        if isinstance(other, ConfigSimulator):
            if self._dict == other._dict:
                return True
        return False

    def __str__(self):
        return str(self.dict())

    def __repr__(self):
        return (
            "ConfigSimulator(number_of_sidebands={0}, integration_volume={1}, "
            "integration_density={2}, decompose={3})"
        ).format(
            self.number_of_sidebands,
            self.integration_volume,
            self.integration_density,
            self.decompose,
        )

    # dict
    def dict(self):
        dict_self = deepcopy(self._dict)
        dict_self["integration_volume"] = self.integration_volume
        return dict_self



cimport test as clib

cimport numpy as np
import numpy as np
import cython

from libcpp cimport bool as bool_t


__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


## wigner matrices

@cython.boundscheck(False)
@cython.wraparound(False)
def wigner_d_matrices(int l, np.ndarray[double] angle):
    cdef int n = angle.size
    cdef int n1 = (2*l+1)**2
    cdef np.ndarray[double] wigner = np.empty(n*n1, dtype=np.float64)
    clib.wigner_d_matrices(l, n, &angle[0], &wigner[0])
    return wigner


@cython.boundscheck(False)
@cython.wraparound(False)
def wigner_d_matrices_from_exp_I_beta(int l, np.ndarray[double complex] exp_I_beta):
    r"""
    Returns a :math:`(2l+1) \times (2l+1)` wigner-d(beta) matrix of rank $l$ at
    a given angle `beta` in the form of `exp(i\beta)`. Currently only rank l=2 and
    l=4 is supported.

    If `exp_I_beta` is a 1D-numpy array of size n, a
    `n x (2l+1) x (2l+1)` matrix is returned instead.

    :ivar l: The angular momentum quantum number.
    :ivar exp_I_beta: An 1D numpy array or a scalar representing $\exp\beta$.
    """
    n1 = (2 * l + 1)
    cdef int n = exp_I_beta.size
    cdef np.ndarray[double, ndim=1] wigner = np.empty(n * n1**2)
    clib.wigner_d_matrices_from_exp_I_beta(l, n, &exp_I_beta[0], &wigner[0])
    return wigner.reshape(n, n1, n1)


@cython.boundscheck(False)
@cython.wraparound(False)
def wigner_dm0_vector(int l, double beta):
    r"""

    """
    cdef int n1 = (2 * l + 1)
    cdef np.ndarray[double] R_out = np.zeros(n1, dtype=np.float64)
    clib.wigner_dm0_vector(l, beta, &R_out[0])
    return R_out


## wigner rotations

@cython.boundscheck(False)
@cython.wraparound(False)
def single_wigner_rotation(int l, np.ndarray[double] euler_angles, np.ndarray[double complex] R_in):
    cdef int n1 = (2 * l + 1)
    cdef np.ndarray[double complex] R_out = np.zeros(n1, dtype=np.complex128)
    clib.single_wigner_rotation(l, &euler_angles[0],
                            &R_in[0], &R_out[0])
    return R_out


# @cython.boundscheck(False)
# @cython.wraparound(False)
# def wigner_rotation(int l, np.ndarray[double complex] R_in,
#                     cos_alpha = None, cos_beta = None,
#                     wigner_matrix=None):
#     r"""

#     """
#     cdef int n1 = 2 * l + 1
#     cdef np.ndarray[double, ndim=1] wigner, cos_alpha_c, cos_beta_c
#     cos_alpha_c = np.asarray(cos_alpha, dtype=np.float64)

#     if wigner_matrix is None:
#         n = cos_beta.size
#         wigner = np.empty(n1**2 * n)
#         cos_beta_c = np.asarray(cos_beta, dtype=np.float64)
#         clib.wigner_d_matrices_from_cosines(l, n, &cos_beta_c[0], &wigner[0])
#     else:
#         n = wigner_matrix.shape[0]
#         wigner = np.asarray(wigner_matrix.ravel(), dtype=np.float64)

#     cdef np.ndarray[complex] R_out = np.zeros(n1*n, dtype=np.complex128)

#     clib.__wigner_rotation(l, n, &wigner[0],
#                            &cos_alpha_c[0], &R_in[0], &R_out[0])
#     return R_out.reshape(n, n1)


@cython.boundscheck(False)
@cython.wraparound(False)
def __wigner_rotation_2(int l, np.ndarray[double] cos_alpha,
                        np.ndarray[double] cos_beta,
                        np.ndarray[double complex] R_in):

    cdef int n1 = 2 * l + 1
    cdef int n = cos_alpha.size
    cdef np.ndarray[double, ndim=1] wigner
    cdef np.ndarray[double complex, ndim=1] exp_I_beta
    wigner = np.empty(n1**2 * n, dtype=np.float64)
    sin_beta = np.sqrt(1 - cos_beta**2)
    exp_I_beta = np.asarray(cos_beta + 1j*sin_beta, dtype=np.complex128)
    clib.wigner_d_matrices_from_exp_I_beta(l, n, &exp_I_beta[0], &wigner[0])

    cdef np.ndarray[double complex] exp_im_alpha
    exp_im_alpha = np.empty(4 * n, dtype=np.complex128)
    exp_im_alpha[3*n:] = cos_alpha + 1j*np.sqrt(1.0 - cos_alpha**2)
    clib.get_exp_Im_alpha(n, 1, &exp_im_alpha[0])

    cdef np.ndarray[complex] R_out = np.zeros(n1*n, dtype=np.complex128)


    clib.__wigner_rotation_2(l, n, &wigner[0],
                           &exp_im_alpha[0], &R_in[0], &R_out[0])
    return R_out.reshape(n, n1)



@cython.boundscheck(False)
@cython.wraparound(False)
def get_exp_Im_alpha(int n, np.ndarray[double] cos_alpha, bool_t allow_fourth_rank):
    cdef unsigned int n_ = n
    cdef np.ndarray[double complex] exp_Im_alpha = np.empty(4*n, dtype=np.complex128)
    exp_Im_alpha[3*n:] = cos_alpha + 1j*np.sqrt(1.0 - cos_alpha**2)
    clib.get_exp_Im_alpha(n_, allow_fourth_rank, &exp_Im_alpha[0])
    return exp_Im_alpha


@cython.boundscheck(False)
@cython.wraparound(False)
def pre_phase_components(int number_of_sidebands, double sample_rotation_frequency_in_Hz):
    r"""

    """
    cdef int n1 = 9 * number_of_sidebands
    cdef np.ndarray[double] pre_phase = np.zeros(2*n1, dtype=np.float64)
    clib.__get_components(number_of_sidebands, sample_rotation_frequency_in_Hz, &pre_phase[0])
    return pre_phase.view(dtype=np.complex128).reshape(9, number_of_sidebands)










@cython.boundscheck(False)
@cython.wraparound(False)
def cosine_of_polar_angles_and_amplitudes(int geodesic_polyhedron_frequency=72):
    r"""
    Calculate the direction cosines and the related amplitudes for
    the positive quadrant of the sphere. The direction cosines corresponds to
    angle $\alpha$ and $\beta$, where $\alpha$ is the azimuthal angle and
    $\beta$ is the polar angle. The amplitudes are evaluated as

        `amp = 1/r**3`

    where `r` is the distance from the origin to the face of the unit
    octahedron in the positive quadrant along the line given by the values of
    $\alpha$ and $\beta$.

    :ivar geodesic_polyhedron_frequency:
        The value is an integer which represents the frequency of class I
        geodesic polyhedra. These polyhedra are used in calculating the
        spherical average. Presently we only use octahedral as the frequency1
        polyhedra. As the frequency of the geodesic polyhedron increases, the
        polyhedra approach a sphere geometry. For line-shape simulation, a higher
        frequency will result in a better powder averaging.
        The default value is 72.
        Read more on the `Geodesic polyhedron
        <https://en.wikipedia.org/wiki/Geodesic_polyhedron>`_.

    :return cos_alpha: The cosine of the azimuthal angle.
    :return cos_beta: The cosine of the polar angle.
    :return amp: The amplitude at the given $\alpha$ and $\beta$.
    """
    nt = geodesic_polyhedron_frequency
    cdef unsigned int octant_orientations = int((nt+1) * (nt+2)/2)

    cdef np.ndarray[double complex] exp_I_alpha = np.empty(octant_orientations, dtype=np.complex128)
    cdef np.ndarray[double complex] exp_I_beta = np.empty(octant_orientations, dtype=np.complex128)
    cdef np.ndarray[double] amp = np.empty(octant_orientations, dtype=np.float64)

    clib.octahedron_averaging_setup(nt, &exp_I_alpha[0], &exp_I_beta[0], &amp[0])

    return exp_I_alpha, exp_I_beta, amp


@cython.boundscheck(False)
@cython.wraparound(False)
def octahedronInterpolation(np.ndarray[double] spec, np.ndarray[double, ndim=2] freq, int nt, np.ndarray[double, ndim=2] amp, int stride=1):
    cdef int i
    cdef int number_of_sidebands = amp.shape[0]
    for i in range(number_of_sidebands):
        clib.octahedronInterpolation(&spec[0], &freq[i,0], nt, &amp[i,0], stride, spec.size)


@cython.boundscheck(False)
@cython.wraparound(False)
def triangle_interpolation(vector, np.ndarray[double, ndim=1] spectrum_amp,
                           double amp=1):
    r"""
    Given a vector of three points, this method interpolates the
    between the points to form a triangle. The height of the triangle is given
    as `2.0/(f[2]-f[1])` where `f` is the array `vector` sorted in an ascending
    order.

    :ivar vector: 1-D array of three points.
    :ivar spectrum_amp: A numpy array of amplitudes. This array is updated.
    :ivar offset: A float specifying the offset. The points from array `vector`
                  are incremented or decremented based in this values. The
                  default value is 0.
    :ivar amp: A float specifying the offset. The points from array `vector`
               are incremented or decremented based in this values. The
               default value is 0.
    """
    cdef np.ndarray[int, ndim=1] points = np.asarray([spectrum_amp.size], dtype=np.int32)
    cdef np.ndarray[double, ndim=1] f_vector = np.asarray(vector, dtype=np.float64)

    cdef double *f1 = &f_vector[0]
    cdef double *f2 = &f_vector[1]
    cdef double *f3 = &f_vector[2]

    cdef np.ndarray[double, ndim=1] amp_ = np.asarray([amp])

    clib.triangle_interpolation(f1, f2, f3, &amp_[0], &spectrum_amp[0], &points[0])


@cython.boundscheck(False)
@cython.wraparound(False)
def __batch_wigner_rotation(unsigned int octant_orientations,
                            unsigned int n_octants,
                            np.ndarray[double] wigner_2j_matrices,
                            np.ndarray[double complex] R2,
                            np.ndarray[double] wigner_4j_matrices,
                            np.ndarray[double complex] R4,
                            np.ndarray[double complex] exp_Im_alpha):

    cdef np.ndarray[double complex] w2 = np.empty(5*octant_orientations*n_octants, dtype=np.complex128)
    cdef np.ndarray[double complex] w4 = np.empty(9*octant_orientations*n_octants, dtype=np.complex128)
    clib.__batch_wigner_rotation(octant_orientations, n_octants,
                            &wigner_2j_matrices[0], &R2[0], &wigner_4j_matrices[0],
                            &R4[0], &exp_Im_alpha[0], &w2[0], &w4[0])
    return w2, w4


@cython.boundscheck(False)
@cython.wraparound(False)
def _one_d_simulator(
        # spectrum information
        double reference_offset,
        double increment,
        int number_of_points,

        float spin_quantum_number = 0.5,
        float larmor_frequency = 0.0,

        # CSA tensor information
        isotropic_chemical_shift = None,
        shielding_anisotropy = None,
        shielding_asymmetry = None,
        shielding_orientations = None,

        # quad tensor information
        quadrupolar_coupling_constant = None,
        quadrupole_asymmetry = None,
        quadrupole_orientations = None,

        second_order_quad = 1,
        remove_second_order_quad_isotropic = 0,

        # dipolar coupling
        D = None,

        # spin rate, spin angle and number spinning sidebands
        int number_of_sidebands = 128,
        double sample_rotation_frequency_in_Hz = 0.0,
        rotor_angle = None,

        m_final = 0.5,
        m_initial = -0.5,

        # Euler angle -> principal to molecular frame
        # omega_PM=None,

        # Euler angles for powder averaging scheme
        int geodesic_polyhedron_frequency=90,
        int integration_volume=0):

    nt = geodesic_polyhedron_frequency
    if isotropic_chemical_shift is None:
        isotropic_chemical_shift = 0
    isotropic_chemical_shift = np.asarray([isotropic_chemical_shift], dtype=np.float64).ravel()
    cdef number_of_sites = isotropic_chemical_shift.size
    cdef np.ndarray[double, ndim=1] isotropic_chemical_shift_c = isotropic_chemical_shift

    if spin_quantum_number > 0.5 and larmor_frequency == 0.0:
        raise Exception("'larmor_frequency' is required for quadrupole spins.")

    # Shielding anisotropic values
    if shielding_anisotropy is None:
        shielding_anisotropy = np.ones(number_of_sites, dtype=np.float64).ravel() #*1e-4*increment
    else:
        shielding_anisotropy = np.asarray([shielding_anisotropy], dtype=np.float64).ravel()
    if shielding_anisotropy.size != number_of_sites:
        raise Exception("Number of shielding anisotropies are not consistent with the number of spins.")
    cdef np.ndarray[double, ndim=1] shielding_anisotropy_c = shielding_anisotropy

    # Shielding asymmetry values
    if shielding_asymmetry is None:
        shielding_asymmetry = np.zeros(number_of_sites, dtype=np.float64).ravel()
    else:
        shielding_asymmetry = np.asarray([shielding_asymmetry], dtype=np.float64).ravel()
    if shielding_asymmetry.size != number_of_sites:
        raise Exception("Number of shielding asymmetry are not consistent with the number of spins.")
    cdef np.ndarray[double, ndim=1] shielding_asymmetry_c = shielding_asymmetry

    # Shielding orientations
    if shielding_orientations is None:
        shielding_orientations = np.zeros(3*number_of_sites, dtype=np.float64).ravel()
    else:
        shielding_orientations = np.asarray([shielding_orientations], dtype=np.float64).ravel()
    if shielding_orientations.size != 3*number_of_sites:
        raise Exception("Number of euler angles are not consistent with the number of shielding tensors.")
    cdef np.ndarray[double, ndim=1] shielding_orientations_c = shielding_orientations*np.pi/180.0

    # Quad coupling constant
    if quadrupolar_coupling_constant is None:
        quadrupolar_coupling_constant = np.zeros(number_of_sites, dtype=np.float64).ravel()
    else:
        quadrupolar_coupling_constant = np.asarray([quadrupolar_coupling_constant], dtype=np.float64).ravel()
    if quadrupolar_coupling_constant.size != number_of_sites:
        raise Exception("Number of quad coupling constants are not consistent with the number of spins.")
    cdef np.ndarray[double, ndim=1] quadrupolar_coupling_constant_c = quadrupolar_coupling_constant

    # Quad asymmetry value
    if quadrupole_asymmetry is None:
        quadrupole_asymmetry = np.zeros(number_of_sites, dtype=np.float64).ravel()
    else:
        quadrupole_asymmetry = np.asarray([quadrupole_asymmetry], dtype=np.float64).ravel()
    if quadrupole_asymmetry.size != number_of_sites:
        raise Exception("Number of quad asymmetry are not consistent with the number of spins.")
    cdef np.ndarray[double, ndim=1] quadrupole_asymmetry_c = quadrupole_asymmetry

    # Quadrupolar orientations
    if quadrupole_orientations is None:
        quadrupole_orientations = np.zeros(3*number_of_sites, dtype=np.float64).ravel()
    else:
        quadrupole_orientations = np.asarray([quadrupole_orientations], dtype=np.float64).ravel()
    if quadrupole_orientations.size != 3*number_of_sites:
        raise Exception("Number of euler angles are not consistent with the number of quad tensors.")
    cdef np.ndarray[double, ndim=1] quadrupole_orientations_c = quadrupole_orientations*np.pi/180.0

    # Dipolar coupling constant
    if D is None:
        D = np.zeros(number_of_sites, dtype=np.float64).ravel()
    else:
        D = np.asarray([D], dtype=np.float64).ravel()
    if D.size != number_of_sites:
        raise Exception("Number of dipolar coupling are not consistent with the number of spins.")
    cdef np.ndarray[double, ndim=1] D_c = D

    if rotor_angle is None:
        rotor_angle = 54.735
    cdef double rotor_angle_in_rad_c = np.pi*rotor_angle/180.

    cdef second_order_quad_c = second_order_quad

    cdef np.ndarray[double, ndim=1] transition_c = np.asarray([m_initial, m_final], dtype=np.float64)

    cdef np.ndarray[double, ndim=1] amp = np.zeros(number_of_points * number_of_sites)

    cdef clib.isotopomer_ravel isotopomer_struct

    isotopomer_struct.number_of_sites = number_of_sites
    isotopomer_struct.spin = spin_quantum_number
    isotopomer_struct.larmor_frequency = larmor_frequency

    isotopomer_struct.isotropic_chemical_shift_in_Hz = &isotropic_chemical_shift_c[0]
    isotopomer_struct.shielding_anisotropy_in_Hz = &shielding_anisotropy_c[0]
    isotopomer_struct.shielding_asymmetry = &shielding_asymmetry_c[0]
    isotopomer_struct.shielding_orientation = &shielding_orientations_c[0]

    isotopomer_struct.quadrupole_coupling_constant_in_Hz = &quadrupolar_coupling_constant_c[0]
    isotopomer_struct.quadrupole_asymmetry = &quadrupole_asymmetry_c[0]
    isotopomer_struct.quadrupole_orientation = &quadrupole_orientations_c[0]

    isotopomer_struct.dipolar_couplings = &D_c[0]

    cdef int remove_second_order_quad_isotropic_c = remove_second_order_quad_isotropic
    clib.spinning_sideband_core(
            # spectrum information and related amplitude
            &amp[0],
            reference_offset,
            increment,
            number_of_points,

            &isotopomer_struct,

            second_order_quad_c,
            remove_second_order_quad_isotropic_c,

            # spin rate, spin angle and number spinning sidebands
            number_of_sidebands,
            sample_rotation_frequency_in_Hz,
            rotor_angle_in_rad_c,

            &transition_c[0],
            geodesic_polyhedron_frequency,
            integration_volume             # 0-octant, 1-hemisphere, 2-sphere.
            )


    freq = np.arange(number_of_points)*increment + reference_offset

    return freq, amp



# -*- coding: utf-8 -*-
"""Tests for the base Parseable pattern"""
from typing import ClassVar

import pytest
from mrsimulator.parseable import enforce_units
from mrsimulator.parseable import Parseable


class ParseableTestClass(Parseable):
    """
    Dummy test class for Parseable pattern
    """

    foo: float = 0
    bar: int = 0
    property_unit_types: ClassVar = {
        "foo": "angle",
        "bar": ["dimensionless", "frequency"],
    }
    property_default_units: ClassVar = {"foo": "rad", "bar": ["pct", "Hz"]}


# Test Enforce Units
def test_good_units():
    enforce_units("300 Hz", "frequency", "Hz")


def test_bad_units():
    with pytest.raises(Exception):
        enforce_units("300 Hz", "angle", "rad")


# Test Parseable pattern


def test_blank_init():
    pr = ParseableTestClass()
    pr


def test_parse_json():

    good_json = {"foo": "300 rad", "bar": "300 ppm"}
    pr = ParseableTestClass.parse_dict_with_units(good_json)
    pr

    good_json2 = {"foo": "300 rad", "bar": "300 Hz"}
    pr = ParseableTestClass.parse_dict_with_units(good_json2)
    pr

    bad_json = {"foo": "300 Hz", "bar": "300 ppm"}

    with pytest.raises(Exception) as err:
        ParseableTestClass.parse_dict_with_units(bad_json)
    assert (
        str(err.value) == "Error enforcing units for foo: 300 Hz\n"
        "A angle value is required but got a frequency instead"
    )



# -*- coding: utf-8 -*-
__version__ = "0.2.0"

from mrsimulator.parseable import Parseable
from mrsimulator.tensors import SymmetricTensor, AntisymmetricTensor
from mrsimulator.site import Site
from mrsimulator.dimension import Dimension
from mrsimulator.isotopomer import Isotopomer
from mrsimulator.simulator import Simulator



# -*- coding: utf-8 -*-
"""Test for the base Site class."""
# import os.path
import pytest
from mrsimulator import Site
from pydantic import ValidationError


# Test Site ===========================================================================
def test_direct_init_site1():
    # test 1 --------------------------------------------------------------------------
    the_site = Site(isotope="29Si", isotropic_chemical_shift=10)
    assert the_site.isotope == "29Si"
    assert the_site.isotropic_chemical_shift == 10
    assert the_site.property_units["isotropic_chemical_shift"] == "ppm"

    assert the_site.shielding_antisymmetric is None
    assert the_site.quadrupolar is None
    assert the_site.shielding_symmetric is None

    assert the_site.atomic_number == 14
    assert the_site.gyromagnetic_ratio == -8.465499
    assert the_site.natural_abundance == 4.683
    assert the_site.quadrupole_moment == 0.0
    assert the_site.spin == 0.5

    # test 2 --------------------------------------------------------------------------
    the_site = Site(
        isotope="29Si",
        isotropic_chemical_shift=10,
        shielding_symmetric={"zeta": 12.1, "eta": 0.1},
    )
    assert the_site.isotope == "29Si"
    assert the_site.isotropic_chemical_shift == 10
    assert the_site.property_units["isotropic_chemical_shift"] == "ppm"

    assert the_site.shielding_antisymmetric is None

    assert the_site.quadrupolar is None

    assert the_site.shielding_symmetric.zeta == 12.1
    assert the_site.shielding_symmetric.property_units["zeta"] == "ppm"
    assert the_site.shielding_symmetric.eta == 0.1

    # site properties
    assert the_site.atomic_number == 14
    assert the_site.gyromagnetic_ratio == -8.465499
    assert the_site.natural_abundance == 4.683
    assert the_site.quadrupole_moment == 0.0
    assert the_site.spin == 0.5

    # test 3 --------------------------------------------------------------------------
    error = "ensure this value is less than or equal to 1"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        Site(
            isotope="29Si",
            isotropic_chemical_shift=10,
            shielding_symmetric={"zeta": 12.1, "eta": 1.5},
        )

    assert Site().isotope == "1H"

    error = ["with spin quantum number", "does not allow quadrupolar tensor"]
    with pytest.raises(ValidationError, match=".*{0}.*{1}.*".format(*error)):
        Site(quadrupolar={"Cq": 5.1e6})

    with pytest.raises(ValidationError, match=".*{0}.*{1}.*".format(*error)):
        Site.parse_dict_with_units(dict(quadrupolar={"Cq": "5.1 MHz"}))


def test_parse_json_site():
    # test 1 --------------------------------------------------------------------------
    good_json_site = {
        "isotope": "1H",
        "isotropic_chemical_shift": "0 ppm",
        "shielding_symmetric": {"zeta": "13.89 ppm", "eta": 0.25},
    }

    # testing method parse_dict_with_units()
    the_site = Site.parse_dict_with_units(good_json_site)
    assert the_site.isotope == "1H"
    assert the_site.isotropic_chemical_shift == 0
    assert the_site.property_units["isotropic_chemical_shift"] == "ppm"

    assert the_site.shielding_antisymmetric is None
    assert the_site.quadrupolar is None

    assert the_site.shielding_symmetric.zeta == 13.89
    assert the_site.shielding_symmetric.property_units["zeta"] == "ppm"
    assert the_site.shielding_symmetric.eta == 0.25
    assert the_site.shielding_symmetric.alpha is None
    assert the_site.shielding_symmetric.beta is None
    assert the_site.shielding_symmetric.gamma is None

    # site properties
    assert the_site.atomic_number == 1
    assert the_site.gyromagnetic_ratio == 42.57748
    assert the_site.natural_abundance == 99.985
    assert the_site.quadrupole_moment == 0.0
    assert the_site.spin == 0.5

    # test 2 --------------------------------------------------------------------------
    good_json_2 = {
        "isotope": "14N",
        "isotropic_chemical_shift": "-10 ppm",
        "quadrupolar": {"Cq": "5.12 MHz", "eta": 0.5},
    }

    the_site = Site.parse_dict_with_units(good_json_2)
    assert the_site.isotope == "14N"
    assert the_site.isotropic_chemical_shift == -10
    assert the_site.property_units["isotropic_chemical_shift"] == "ppm"

    assert the_site.shielding_antisymmetric is None
    assert the_site.shielding_symmetric is None

    assert the_site.quadrupolar.Cq == 5120000.0
    assert the_site.quadrupolar.eta == 0.5

    # site properties
    assert the_site.atomic_number == 7
    assert the_site.gyromagnetic_ratio == 3.077706
    assert the_site.natural_abundance == 99.634
    assert the_site.quadrupole_moment == 0.0193
    assert the_site.spin == 1

    # test 3 bad input ----------------------------------------------------------------
    bad_json = {"isotope": "1H", "isotropic_chemical_shift": "0 rad"}

    with pytest.raises(Exception):
        Site.parse_dict_with_units(bad_json)

    error = "Error enforcing units for isotropic_chemical_shift: 10 MHz"
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        Site.parse_dict_with_units(
            {
                "isotope": "29Si",
                "isotropic_chemical_shift": "10 MHz",
                "shielding_symmetric": {"zeta": "12.1 ppm", "eta": 0.5},
            }
        )


def test_site_object_methods():
    good_json_2 = {"isotope": "14N", "isotropic_chemical_shift": "-10 ppm"}
    the_site = Site.parse_dict_with_units(good_json_2)

    # testing method dict()
    result = {
        "isotope": "14N",
        "isotropic_chemical_shift": -10.0,
        "property_units": {"isotropic_chemical_shift": "ppm"},
        "name": None,
        "quadrupolar": None,
        "shielding_symmetric": None,
        "shielding_antisymmetric": None,
    }
    assert the_site.dict() == result, "Failed Site.dict()"

    # testing method to_freq_dict()
    result = {
        "isotope": "14N",
        "isotropic_chemical_shift": -1 * 3.077706 * 9.4 * -10.0,  # -gamma * B0 * iso
        "name": None,
        "quadrupolar": None,
        "shielding_symmetric": None,
        "shielding_antisymmetric": None,
    }
    assert the_site.to_freq_dict(B0=9.4) == result, "Failed Site.to_freq_dict()"

    # testing method to_dict_with_units()
    result = {"isotope": "14N", "isotropic_chemical_shift": "-10.0 ppm"}
    assert the_site.to_dict_with_units() == result, "Failed Site.to_dict_with_units()"

    result = {
        "isotope": "27Al",
        "isotropic_chemical_shift": "10.0 ppm",
        "shielding_symmetric": {"zeta": "12.1 ppm", "eta": 0.1, "alpha": "2.1 rad"},
        "shielding_antisymmetric": {
            "zeta": "-1.1 ppm",
            "alpha": "0.1 rad",
            "beta": "2.5 rad",
        },
        "quadrupolar": {"Cq": "10000000.0 Hz", "eta": 0.6},
    }
    the_site = Site(
        isotope="27Al",
        isotropic_chemical_shift=10,
        shielding_symmetric={"zeta": 12.1, "eta": 0.1, "alpha": 2.1},
        shielding_antisymmetric={"zeta": -1.1, "alpha": 0.1, "beta": 2.5},
        quadrupolar={"Cq": 10e6, "eta": 0.6},
    )
    assert the_site.to_dict_with_units() == result, "Failed Site.to_dict_with_units()"

    larmor_freq = -1 * 11.10309 * 9.4  # -gamma * B0
    result = {
        "name": None,
        "isotope": "27Al",
        "isotropic_chemical_shift": 10.0 * larmor_freq,  # larmor_freq * iso
        "shielding_symmetric": {
            "zeta": 12.1 * larmor_freq,
            "eta": 0.1,
            "alpha": 2.1,
            "beta": None,
            "gamma": None,
        },
        "shielding_antisymmetric": {
            "zeta": -1.1 * larmor_freq,
            "alpha": 0.1,
            "beta": 2.5,
        },
        "quadrupolar": {
            "Cq": 10000000.0,
            "zeta": None,
            "eta": 0.6,
            "alpha": None,
            "beta": None,
            "gamma": None,
        },
    }
    assert the_site.to_freq_dict(9.4) == result, "Failed Site.to_dict_with_units()"


def test_equality():
    a = Site(isotope="1H")
    b = Site(isotope="1H")
    assert a == b

    c = Site(isotope="1H", isotropic_chemical_shift=16)
    assert a != c



# -*- coding: utf-8 -*-
"""Test for the base Isotopomers class."""
from random import randint

import pytest
from mrsimulator import Isotopomer
from mrsimulator import Site
from mrsimulator.isotopomer import allowed_isotopes
from pydantic import ValidationError


def test_direct_init_isotopomer():
    the_isotopomer = Isotopomer(sites=[], abundance=10)

    assert the_isotopomer.sites == []
    assert the_isotopomer.abundance == 10.0

    test_site = Site(isotope="29Si", isotropic_chemical_shift=10)

    assert test_site.isotope == "29Si"
    assert test_site.isotropic_chemical_shift == 10.0
    assert test_site.property_units["isotropic_chemical_shift"] == "ppm"

    the_isotopomer = Isotopomer(sites=[test_site], abundance=10)
    assert isinstance(the_isotopomer.sites[0], Site)
    assert the_isotopomer.abundance == 10.0

    the_isotopomer = Isotopomer(sites=[test_site, test_site], abundance=10)
    assert isinstance(the_isotopomer.sites[0], Site)
    assert isinstance(the_isotopomer.sites[1], Site)
    assert id(the_isotopomer.sites[0]) != id(the_isotopomer.sites[1])
    assert the_isotopomer.abundance == 10.0

    the_isotopomer = Isotopomer(
        name="Just a test",
        description="The same",
        sites=[
            {"isotope": "1H", "isotropic_chemical_shift": 0},
            {
                "isotope": "17O",
                "isotropic_chemical_shift": -10,
                "quadrupolar": {"Cq": 5.1e6, "eta": 0.5},
            },
        ],
        abundance=4.23,
    )
    assert the_isotopomer.name == "Just a test"
    assert the_isotopomer.description == "The same"
    assert the_isotopomer.sites[0].isotope == "1H"
    assert the_isotopomer.sites[0].isotropic_chemical_shift == 0
    assert the_isotopomer.sites[1].isotope == "17O"
    assert the_isotopomer.sites[1].isotropic_chemical_shift == -10
    assert the_isotopomer.sites[1].quadrupolar.Cq == 5.1e6
    assert the_isotopomer.sites[1].quadrupolar.eta == 0.5
    assert the_isotopomer.abundance == 4.23


def test_parse_json_isotopomer():
    good_json = {"sites": [], "abundance": "10%"}

    good_json2 = {
        "sites": [{"isotope": "1H", "isotropic_chemical_shift": "0 ppm"}],
        "abundance": "10%",
    }

    iso1 = Isotopomer.parse_dict_with_units(good_json)
    assert len(iso1.sites) == 0
    assert iso1.abundance == 10

    iso2 = Isotopomer.parse_dict_with_units(good_json2)
    assert len(iso2.sites) == 1
    assert iso2.sites[0].isotope == "1H"
    assert iso2.sites[0].isotropic_chemical_shift == 0
    assert iso2.abundance == 10

    bad_json = {"sites": [], "abundance": "10 Hz"}
    with pytest.raises(Exception):
        Isotopomer.parse_dict_with_units(bad_json)


def test_isotopomer_methods():
    good_json2 = {
        "sites": [{"isotope": "1H", "isotropic_chemical_shift": "2 ppm"}],
        "abundance": "10%",
    }

    # to_freq_dict()
    iso1 = Isotopomer.parse_dict_with_units(good_json2).to_freq_dict(9.4)
    result = {
        "name": "",
        "description": "",
        "sites": [
            {
                "isotope": "1H",
                "isotropic_chemical_shift": -2 * 42.57748 * 9.4,  # -gamma * B0 * iso
                "name": None,
                "quadrupolar": None,
                "shielding_antisymmetric": None,
                "shielding_symmetric": None,
            }
        ],
        "abundance": 10,
    }
    assert iso1 == result

    # to_dict_with_units()
    iso1 = Isotopomer.parse_dict_with_units(good_json2).to_dict_with_units()
    result = {
        "name": "",
        "description": "",
        "sites": [{"isotope": "1H", "isotropic_chemical_shift": "2.0 ppm"}],
        "abundance": "10.0%",
    }
    assert iso1 == result


def get_isotopomer_list():
    isotopes = ["19F", "31P", "2H", "6Li", "14N", "27Al", "25Mg", "45Sc", "87Sr"]
    sites = []
    for isotope in isotopes:
        for _ in range(randint(1, 3)):
            sites.append(Site(isotope=isotope))
    return Isotopomer(sites=sites)


def test_get_isotopes():
    isotopes = {"19F", "31P", "2H", "6Li", "14N", "27Al", "25Mg", "45Sc", "87Sr"}
    isotopomer = get_isotopomer_list()
    assert isotopomer.get_isotopes() == isotopes
    assert isotopomer.get_isotopes(I=0.5) == {"19F", "31P"}
    assert isotopomer.get_isotopes(I=1) == {"2H", "6Li", "14N"}
    assert isotopomer.get_isotopes(I=1.5) == set()
    assert isotopomer.get_isotopes(I=2.5) == {"27Al", "25Mg"}
    assert isotopomer.get_isotopes(I=3.5) == {"45Sc"}
    assert isotopomer.get_isotopes(I=4.5) == {"87Sr"}


def test_allowed_isotopes():
    assert {"19F", "31P", "129Xe", "1H", "57Fe", "13C", "15N", "29Si"}.issubset(
        set(allowed_isotopes(I=0.5))
    )
    assert {"2H", "6Li", "14N"}.issubset(set(allowed_isotopes(I=1)))
    assert {"7Li", "9Be", "11B", "21Ne", "23Na", "33S", "37Cl", "41K"}.issubset(
        set(allowed_isotopes(I=1.5))
    )
    assert {"17O", "25Mg", "27Al", "47Ti", "55Mn", "67Zn"}.issubset(
        set(allowed_isotopes(I=2.5))
    )
    assert {"43Ca", "45Sc", "49Ti", "51V", "59Co", "123Sb", "133Cs"}.issubset(
        set(allowed_isotopes(I=3.5))
    )
    assert {"73Ge", "83Kr", "87Sr", "93Nb", "113In"}.issubset(
        set(allowed_isotopes(I=4.5))
    )


def test_bad_assignments():
    error = "value is not a valid list"
    with pytest.raises(ValidationError, match=".*{0}.*".format(error)):
        Isotopomer(sites=Site())


def test_equality():
    a = Isotopomer(sites=[Site(isotope="1H")])
    b = Isotopomer(sites=[Site(isotope="1H")])
    assert a == b

    c = Isotopomer(sites=[Site(isotope="1H", isotropic_chemical_shift=16)])
    assert a != c



# -*- coding: utf-8 -*-
"""Base Dimension class."""
import os.path
import re
import warnings
from typing import ClassVar
from typing import Dict
from typing import Optional

import numpy as np
from monty.serialization import loadfn
from mrsimulator import Parseable
from pydantic import Field
from pydantic import validator

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
ISOTOPE_DATA = loadfn(os.path.join(MODULE_DIR, "isotope_data.json"))


class Dimension(Parseable):
    r"""
    Base class for an NMR spectroscopic dimension.

    Args:
        number_of_points: An optional integer with the number of points, :math:`N`,
                along the dimension. The default value is ``1024``.
        spectral_width: A `required` float with the spectral width,
                :math:`\Delta x`, along the dimension in units of Hz.
        reference_offset: An `optional` float with the reference offset, :math:`x_0`
                along the dimension in units of Hz. The default value is ``0``.
        magnetic_flux_density: An `optional` float with the macroscopic magnetic flux
                density, :math:`B_0`, of the applied external magnetic field in
                units of T. The default value is ``9.4``.
        rotor_frequency: An `optional` float with the sample spinning frequency
                :math:`\nu_r`, in units of Hz. The default value is ``0``.
        rotor_angle: An `optional` float with the angle between the
                sample rotation axis and the applied external magnetic field,
                :math:`\theta`, in units of rad. The default value is ``0.9553166``,
                i.e. the magic angle.
        isotope: An `optional` isotope string in "{A}{Symbol}" notation such as
                1H or 29Si. The default is`` None``.
        label: An `optional` string label. The default is an empty string.

    Example:
        >>> dim = Dimension(isotope='27Al', spectral_width=50000, rotor_frequency=12000)

    """

    # public attributes
    number_of_points: int = Field(1024, gt=0)
    spectral_width: float = Field(..., gt=0)
    reference_offset: Optional[float] = Field(default=0)
    magnetic_flux_density: Optional[float] = Field(default=9.4, ge=0)
    rotor_frequency: Optional[float] = Field(default=0, ge=0)
    # 54.735 degrees = 0.9553166 radians
    rotor_angle: Optional[float] = Field(default=0.9553166, ge=0, le=1.5707963268)
    isotope: Optional[str] = None
    label: Optional[str] = ""

    property_unit_types: ClassVar = {
        "spectral_width": ["frequency", "dimensionless"],
        "reference_offset": ["frequency", "dimensionless"],
        "magnetic_flux_density": "magnetic flux density",
        "rotor_frequency": "frequency",
        "rotor_angle": "angle",
    }

    property_default_units: ClassVar = {
        "spectral_width": ["Hz", "ppm"],
        "reference_offset": ["Hz", "ppm"],
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    property_units: Dict = {
        "spectral_width": "Hz",
        "reference_offset": "Hz",
        "magnetic_flux_density": "T",
        "rotor_frequency": "Hz",
        "rotor_angle": "rad",
    }

    class Config:
        validate_assignment = True

    @validator("isotope", always=True)
    def get_isotope(cls, v, *, values, **kwargs):
        if v is None:
            return v
        return format_isotope_string(v)

    @property
    def spin(self):
        """
        Spin quantum number, I.

        Example:
            >>> dim.spin
            2.5
        """
        if self.isotope is None:
            return None
        isotope_data = get_isotope_data(self.isotope)
        return isotope_data["spin"] / 2.0

    @property
    def natural_abundance(self):
        """
        Natural abundance of the spin.

        Example:
            >>> dim.natural_abundance
            100.0
        """
        if self.isotope is None:
            return None
        isotope_data = get_isotope_data(self.isotope)
        return isotope_data["natural_abundance"]

    @property
    def gyromagnetic_ratio(self):
        """
        Reduced gyromagnetic ratio of the nucleus given in units of MHz/T.

        Example:
            >>> dim.gyromagnetic_ratio
            11.10309
        """
        if self.isotope is None:
            return None
        isotope_data = get_isotope_data(self.isotope)
        return isotope_data["gyromagnetic_ratio"]

    @property
    def quadrupole_moment(self):
        """
        Quadrupole moment of the nucleus given in units of electron-barn.

        Example:
            >>> dim.quadrupole_moment
            0.15
        """
        if self.isotope is None:
            return None
        isotope_data = get_isotope_data(self.isotope)
        return isotope_data["quadrupole_moment"]

    @property
    def atomic_number(self):
        """
        Atomic number of the nucleus.

        Example:
            >>> dim.atomic_number
            13
        """
        if self.isotope is None:
            return None
        isotope_data = get_isotope_data(self.isotope)
        return isotope_data["atomic_number"]

    @property
    def larmor_frequency(self):
        r"""
        Signed Larmor frequency, :math:`\omega_0=-\gamma B_0`, of the isotope,
        in units of Hz, where :math:`\gamma` is the gyromagnetic ratio of the
        isotope and :math:`B_0` is the macroscopic magnetic flux density
        of the applied external field.

        Example:
            >>> dim.larmor_frequency
            -104369046.0
        """
        if self.isotope is None:
            return None
        return -self.gyromagnetic_ratio * self.magnetic_flux_density * 1e6  # in Hz

    @property
    def coordinates_Hz(self):
        r"""
        The grid coordinates along the dimension in units of Hz, evaluated as

        .. math::
            x_\text{Hz} = \left([0, 1, ... N-1] - T\right) \frac{\Delta x}{N} + x_0

        where :math:`T=N/2` and :math:`T=(N-1)/2` for even and odd values of
        :math:`N`, respectively.

        Example:
            >>> dim.coordinates_Hz[:5]
            array([-25000.      , -24951.171875, -24902.34375 , -24853.515625,
                   -24804.6875  ])
        """
        n = self.number_of_points
        Tk = int(n / 2)
        return (np.arange(n) - Tk) * (self.spectral_width / n) + self.reference_offset

    @property
    def coordinates_ppm(self):
        r"""
        The grid coordinates along the dimension as dimension frequency ratio
        in units of ppm. The coordinates are evaluated as

        .. math::
            x_\text{ppm} = \frac{x_\text{Hz}} {x_0 + \omega_0}

        where :math:`\omega_0` is the Larmor frequency.

        Example:
            >>> dim.coordinates_ppm[:5]
            array([-239.53462217, -239.06678111, -238.59894005, -238.13109899,
                   -237.66325794])
        """
        if self.isotope is None:
            warnings.warn(
                (
                    "The coordinates along the dimension without an assigned "
                    "isotope cannot be converted to dimensionless frequency ratio."
                )
            )
        else:

            denominator = (self.reference_offset + self.larmor_frequency) / 1e6
            return self.coordinates_Hz / abs(denominator)

    @classmethod
    def parse_dict_with_units(cls, py_dict):
        """
        Parse the physical quantities of a Dimension object
        when expressed as a python dictionary.

        Args:
            py_dict: Dict object

        Example:
            >>> dimension_1 = {
            ...     "number_of_points": 1024,
            ...     "spectral_width": "100 Hz",
            ...     "reference_offset": "0 Hz",
            ...     "magnetic_flux_density": "9.4 T",
            ...     "rotor_frequency": "0 Hz",
            ...     "rotor_angle": "54.935 degree",
            ...     "isotope": "29Si",
            ... }
            >>> dimension_object = Dimension.parse_dict_with_units(dimension_1)
        """
        return super().parse_dict_with_units(py_dict)

    def to_dict_with_units(self):
        """
        Serialize the Dimension object to a JSON compliant python dictionary with
        units.

        Example:
            >>> pprint(dimension_object.to_dict_with_units())
            {'isotope': '29Si',
             'label': '',
             'magnetic_flux_density': '9.4 T',
             'number_of_points': 1024,
             'reference_offset': '0.0 Hz',
             'rotor_angle': '0.958796624583085 rad',
             'rotor_frequency': '0.0 Hz',
             'spectral_width': '100.0 Hz'}
        """
        temp_dict = {
            k: v
            for k, v in self.dict(exclude={"property_units"}).items()
            if v is not None
        }
        temp_keys = temp_dict.keys()
        for key, unit in self.property_units.items():
            if key in temp_keys:
                temp_dict[key] = f"{temp_dict[key]} {unit}"
        return temp_dict


def format_isotope_string(isotope_string):
    """Format isotope string to {A}{symbol}, where A is the isotope number"""
    result = re.match(r"(\d+)\s*(\w+)", isotope_string)

    if result is None:
        raise Exception(f"Could not parse isotope string {isotope_string}")
    isotope = result.group(2)
    A = result.group(1)

    formatted_string = f"{A}{isotope}"
    if formatted_string not in ISOTOPE_DATA:
        raise Exception(f"Could not parse isotope string {isotope_string}")

    return formatted_string


def get_isotope_data(isotope_string):
    """
    Gets the isotope's intrinsinc NMR properties from a JSON
    data file.
    """
    formatted_isotope_string = format_isotope_string(isotope_string)
    isotope_dict = dict(ISOTOPE_DATA[formatted_isotope_string])
    isotope_dict.update({"isotope": formatted_isotope_string})
    return isotope_dict



# -*- coding: utf-8 -*-
import numpy as np

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


# def octahedral_coordinate(nt: int):

#     # Do the (x + y + z = nt) face of the octahedron
#     # z -> 0 to nt-1
#     # y -> 0 to nt-z
#     # x -> nt - y - z

#     n = int((nt + 1) * (nt + 2) / 2)
#     x = np.empty(n, dtype=np.float64)
#     y = np.empty(n, dtype=np.float64)
#     z = np.empty(n, dtype=np.float64)

#     k = 0
#     for j in range(nt + 1):
#         for i in range(nt - j + 1):
#             # x = nt-i-j;
#             # y = i;
#             # z = j;
#             x[k] = nt - i - j
#             y[k] = i
#             z[k] = j
#             k += 1

#     return x, y, z


def octahedral_direction_cosine_squares_and_amplitudes(nt: int):

    # Do the (x + y + z = nt) face of the octahedron
    # z -> 0 to nt-1
    # y -> 0 to nt-z
    # x -> nt - y - z

    n = int((nt + 1) * (nt + 2) / 2)
    x = np.empty(n, dtype=np.float64)
    y = np.empty(n, dtype=np.float64)
    z = np.empty(n, dtype=np.float64)

    k = 0
    for j in range(nt + 1):
        for i in range(nt - j + 1):
            # x = nt-i-j;
            # y = i;
            # z = j;
            x[k] = nt - i - j
            y[k] = i
            z[k] = j
            k += 1

    x *= x
    y *= y
    z *= z

    r2 = x + y + z

    x /= r2
    y /= r2
    z /= r2
    amp = nt / (r2 * np.sqrt(r2))

    return x, y, z, amp


def cosine_of_polar_angles_and_amplitudes(geodesic_polyhedron_frequency: int = 72):
    r"""
    Calculates and return the direction cosines and the related amplitudes for
    the positive quadrant of the sphere. The direction cosines corresponds to
    angle $\alpha$ and $\beta$, where $\alpha$ is the azimuthal angle and
    $\beta$ is the polar angle. The amplitudes are evaluated as $\frac{1}{r^3}$
    where $r$ is the distance from the origin to the face of the unit
    octahedron in the positive quadrant along the line given by the values of
    $\alpha$ and $\beta$.

    :ivar geodesic_polyhedron_frequency:
        The value is an integer which represents the frequency of class I
        geodesic polyhedra. These polyhedra are used in calculating the
        spherical average. Presently we only use octahedral as the frequency1
        polyhedra. As the frequency of the geodesic polyhedron increases, the
        polyhedra approach a sphere geometry. For line-shape simulation, a
        higher  frequency will result in a better powder averaging.
        The default value is 72.
        Read more on the `Geodesic polyhedron
        <https://en.wikipedia.org/wiki/Geodesic_polyhedron>`_.

    :return cos_alpha: The cosine of the azimuthal angle.
    :return cos_beta: The cosine of the polar angle.
    :return amp: The amplitude at the given $\alpha$ and $\beta$.
    """
    nt = geodesic_polyhedron_frequency
    xr, yr, zr, amp = octahedral_direction_cosine_squares_and_amplitudes(nt)

    cos_beta = np.sqrt(zr)
    cos_alpha = np.zeros(xr.size, dtype=np.float64)
    cos_alpha[:-1] = np.sqrt(xr[:-1] / (xr[:-1] + yr[:-1]))
    cos_alpha[-1] = 0.0
    return cos_alpha, cos_beta, amp


def triangle_interpolation(f, spec, amp=1.0):
    points = spec.size
    f = np.asarray(f, dtype=np.float64)

    clip_right1 = clip_right2 = False
    clip_left1 = clip_left2 = False

    p = int(f[0])
    # fint = f.astype(int)
    if int(f[0]) == int(f[1]) == int(f[2]):
        if p >= points or p < 0:
            return
        spec[p] += amp
        return

    f = np.sort(f)

    top = amp * 2.0 / (f[2] - f[0])

    p = int(f[0])
    pmid = int(f[1])
    pmax = int(f[2])

    f10 = f[1] - f[0]
    f21 = f[2] - f[1]

    if pmax < 0:
        return

    if p > points:
        return

    if pmid >= points:
        pmid = points
        clip_right1 = True

    if pmax >= points:
        pmax = points
        clip_right2 = True

    if p < 0:
        p = 0
        clip_left1 = True

    if pmid < 0:
        pmid = 0
        clip_left2 = True

    if p != pmid:
        df1 = top / f10
        diff = p + 1.0 - f[0]
        if not clip_left1:
            spec[p] += 0.5 * diff * diff * df1
            p += 1
        else:
            spec[p] += (diff - 0.5) * df1
            p += 1

        diff -= 0.5
        diff *= df1
        while p != pmid:
            diff += df1
            spec[p] += diff
            p += 1
        # spec[p:pmid] += diff + (np.arange(pmid - p, dtype=np.float64) + 1) * df1
        p = pmid
        if not clip_right1:
            spec[p] += (f[1] - p) * (f10 + p - f[0]) * 0.5 * df1

    else:
        if not clip_right1 and not clip_left1:
            spec[p] += f10 * top * 0.5
    if p != pmax:
        df2 = top / f21
        diff = f[2] - p - 1.0

        if not clip_left2:
            spec[p] += (f21 - diff) * (diff + f21) * 0.5 * df2
            p += 1
        else:
            spec[p] += (diff + 0.5) * df2
            p += 1

        diff += 0.5
        diff *= df2
        while p != pmax:
            diff -= df2
            spec[p] += diff
            p += 1
        # spec[p:pmax] += diff - (np.arange(pmax - p, dtype=np.float64) + 1) * df2
        p = pmax
        if not clip_right2:
            spec[p] += (f[2] - p) ** 2 * 0.5 * df2
    else:
        if not clip_right2:
            spec[p] += f21 * top * 0.5


# def average_over_octant(spec, freq, nt, amp):

#     n_pts = (nt + 1) * (nt + 2) / 2

#     #   Interpolate between frequencies by setting up tents

#     local_index = nt - 1
#     i = 0
#     j = 0
#     while i < n_pts - 1:
#         temp = amp[i + 1] + amp[nt + 1 + j]
#         amp1 = temp
#         amp1 += amp[i]

#         freq_3 = [freq[i], freq[i + 1], freq[nt + 1 + j]]
#         triangle_interpolation(freq_3, spec, amp1)

#         if i < local_index:
#             amp1 = temp
#             amp1 += amp[nt + 1 + j + 1]
#             freq_3 = [freq[i + 1], freq[nt + 1 + j], freq[nt + 1 + j + 1]]
#             triangle_interpolation(freq_3, spec, amp1)
#         else:
#             local_index = j + nt
#             i += 1
#         i += 1
#         j += 1



# -*- coding: utf-8 -*-
"""Base Tensor class."""
from typing import ClassVar
from typing import Dict
from typing import Optional

from mrsimulator import Parseable
from pydantic import Field

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


class SymmetricTensor(Parseable):
    """
    Base SymmetricTensor class representing the traceless symmetric part of an
    irreducible second-rank tensor.

    Arguments:
        zeta: The anisotropy parameter of a nuclear shielding tensor expressed using
                Haeberlen convention.
        Cq: The quadrupolar coupling constant derived from an electric field tensor.
        eta: The asymmetry parameter of the SymmetricTensor expressed using
                Haeberlen convention.
        alpha: Euler angle, alpha, given in radian.
        beta: Euler angle, beta, given in radian.
        gamma: Euler angle, gamma, given in radian.
    """

    zeta: Optional[float]
    Cq: Optional[float]
    eta: Optional[float] = Field(default=0, ge=0, le=1)
    alpha: Optional[float]
    beta: Optional[float]
    gamma: Optional[float]

    property_unit_types: ClassVar = {
        "zeta": "dimensionless",
        "Cq": "frequency",
        "alpha": "angle",
        "beta": "angle",
        "gamma": "angle",
    }
    property_default_units: ClassVar = {
        "zeta": "ppm",
        "Cq": "Hz",
        "alpha": "rad",
        "beta": "rad",
        "gamma": "rad",
    }
    property_units: Dict = {
        "zeta": "ppm",
        "Cq": "Hz",
        "alpha": "rad",
        "beta": "rad",
        "gamma": "rad",
    }

    def to_freq_dict(self, larmor_frequency):
        """
        Serialize the SymmetricTensor object to a JSON compliant python dictionary
        where the attribute values are numbers expressed in default units. The default
        unit for attributes with respective dimensionalities are:
        - frequency: `Hz`
        - angle: `rad`

        Args:
            larmor_frequency: The larmor frequency in MHz.

        Return:
            A python dict
        """
        temp_dict = self.dict()
        if temp_dict["zeta"] is not None:
            temp_dict["zeta"] *= larmor_frequency
        temp_dict.pop("property_units")
        return temp_dict

    def to_dict_with_units(self):
        """
        Serialize the SymmetricTensor object to a JSON compliant python dictionary
        with units.

        Return:
            Python dict
        """
        temp_dict = {
            k: f"{v} {self.property_units[k]}"
            for k, v in self.dict().items()
            if v is not None and k not in ["property_units", "eta"]
        }
        if self.eta is not None:
            temp_dict["eta"] = self.eta
        return temp_dict


class AntisymmetricTensor(Parseable):
    """
    Base SymmetricTensor class representing the traceless symmetric part of an
    irreducible second-rank tensor.

    Arguments:
        zeta: The anisotropy parameter of the AntiSymmetricTensor expressed using
                Haeberlen convention.
        alpha: Euler angle, alpha, given in radian.
        beta: Euler angle, beta, given in radian.
    """

    zeta: Optional[float]
    alpha: Optional[float]
    beta: Optional[float]

    property_unit_types: ClassVar = {
        "zeta": "dimensionless",
        "alpha": "angle",
        "beta": "angle",
    }
    property_default_units: ClassVar = {"zeta": "ppm", "alpha": "rad", "beta": "rad"}
    property_units: Dict = {"zeta": "ppm", "alpha": "rad", "beta": "rad"}

    def to_freq_dict(self, larmor_frequency):
        """
        Serialize the AntisymmetricTensor object to a JSON compliant python dictionary
        where the attribute values are numbers expressed in default units. The default
        unit for attributes with respective dimensionalities are:
        - frequency: `Hz`
        - angle: `rad`

        Args:
            larmor_frequency: The larmor frequency in MHz.

        Return:
            Python dict
        """
        temp_dict = self.dict()
        if temp_dict["zeta"] is not None:
            temp_dict["zeta"] *= larmor_frequency
        temp_dict.pop("property_units")
        return temp_dict

    def to_dict_with_units(self):
        """
        Serialize the AntiSymmetricTensor object to a JSON compliant python dictionary
        with units.

        Return:
            Python dict
        """
        temp_dict = {
            k: f"{v} {self.property_units[k]}"
            for k, v in self.dict().items()
            if v is not None and k != "property_units"
        }
        return temp_dict



# -*- coding: utf-8 -*-
"""Utility function module."""
import json
import sys
from os import path
from urllib.parse import urlparse

import requests


__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


def download_file_from_url(url):
    res = urlparse(url)
    filename = path.split(res[2])[1]
    name, extension = path.splitext(filename)
    original_name = name
    i = 0
    while path.isfile(filename):
        filename = "{0}_{1}{2}".format(original_name, str(i), extension)
        i += 1

    with open(filename, "wb") as f:
        response = requests.get(url, stream=True)
        total = response.headers.get("content-length")

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            sys.stdout.write(
                "Downloading '{0}'\nfrom '{1}' to file '{2}'.".format(
                    res[2], res[1], filename
                )
            )
            for data in response.iter_content(
                chunk_size=max(int(total / 1000), 1024 * 1024)
            ):
                downloaded += len(data)
                f.write(data)
                done = int(4 * downloaded / total)
                sys.stdout.write("\n[{}{}]".format("█" * done, "." * (4 - done)))
                sys.stdout.flush()

    sys.stdout.write("\n")

    return filename


def import_json(filename):
    res = urlparse(filename)
    if res[0] not in ["file", ""]:
        filename = download_file_from_url(filename)
    with open(filename, "rb") as f:
        content = f.read()
        return json.loads(str(content, encoding="UTF-8"))



# -*- coding: utf-8 -*-
import platform
import sys
from os.path import abspath
from os.path import dirname
from os.path import join
from os.path import split

import numpy as np
import numpy.distutils.system_info as sysinfo
from Cython.Build import cythonize
from setuptools import Extension
from setuptools import find_packages
from setuptools import setup

# get the version from file

print("python version", sys.version_info)

with open("src/mrsimulator/__init__.py", "r") as f:
    for line in f.readlines():
        if "__version__" in line:
            before_keyword, keyword, after_keyword = line.partition("=")
            version = after_keyword.strip()[1:-1]
            print(version)

module_dir = dirname(abspath(__file__))
print(module_dir)

include_dirs = [
    "/opt/local/include/",
    "/usr/include/",
    "/usr/include/openblas",
    "/usr/include/x86_64-linux-gnu/",
]

library_dirs = [
    "/opt/local/lib/",
    "/usr/lib64/",
    "/usr/lib/",
    "/usr/lib/x86_64-linux-gnu/",
]

libraries = []
data_files = []

numpy_include = np.get_include()

if platform.system() == "Windows":
    conda_location = numpy_include
    for _ in range(5):
        conda_location = split(conda_location)[0]
    include_dirs += [join(conda_location, "Library", "include", "fftw")]
    include_dirs += [join(conda_location, "Library", "include", "openblas")]
    include_dirs += [join(conda_location, "Library", "include")]
    include_dirs += [join(conda_location, "include")]
    library_dirs += [join(conda_location, "Library", "lib")]
    libraries += ["fftw3", "openblas"]
    name = "openblas"

    extra_link_args = ["-lm"]
    extra_compile_args = ["-DFFTW_DLL"]

# this section is important for travis-ci build.
else:
    libraries = ["openblas", "fftw3", "fftw3_threads", "pthread"]
    openblas_info = sysinfo.get_info("openblas")
    fftw3_info = sysinfo.get_info("fftw3")

    if openblas_info != {}:
        name = "openblas"
        library_dirs += openblas_info["library_dirs"]
        libraries += openblas_info["libraries"]
    fftw_keys = fftw3_info.keys()
    if "include_dirs" in fftw_keys:
        include_dirs += fftw3_info["include_dirs"]
    if "library_dirs" in fftw_keys:
        library_dirs += fftw3_info["library_dirs"]
    if "libraries" in fftw_keys:
        libraries += fftw3_info["libraries"]

    extra_link_args = ["-lm"]
    extra_compile_args = ["-g", "-O3"]

include_dirs = list(set(include_dirs))
library_dirs = list(set(library_dirs))
libraries = list(set(libraries))

# other include paths
include_dirs += ["src/c_lib/include/", numpy_include]
print(include_dirs)
print(library_dirs)
print(libraries)

print(extra_compile_args)
print(extra_link_args)

source = [
    "src/c_lib/lib/angular_momentum.c",
    "src/c_lib/lib/interpolation.c",
    "src/c_lib/lib/mrsimulator.c",
    "src/c_lib/lib/octahedron.c",
    "src/c_lib/lib/spinning_sidebands.c",
    "src/c_lib/lib/powder_setup.c",
    "src/c_lib/lib/averaging_scheme.c",
]

# method
ext_modules = [
    Extension(
        name="mrsimulator.methods",
        sources=[*source, "src/c_lib/mrmethods/nmr_methods.pyx"],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        # data_files=data_files,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# tests
ext_modules += [
    Extension(
        name="mrsimulator.tests.tests",
        sources=[*source, "src/c_lib/test/test.pyx"],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

# sandbox
ext_modules += [
    Extension(
        name="mrsimulator.sandbox",
        sources=[*source, "src/c_lib/sandbox/sandbox.pyx"],
        include_dirs=include_dirs,
        language="c",
        libraries=libraries,
        library_dirs=library_dirs,
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
    )
]

setup(
    name="mrsimulator",
    version=version,
    description="A python toolbox for simulating NMR spectra",
    long_description=open(join(module_dir, "README.md")).read(),
    author="Deepansh J. Srivastava",
    author_email="deepansh2012@gmail.com",
    python_requires=">=3.6",
    url="https://github.com/DeepanshS/MRsimulator/",
    packages=find_packages("src"),
    package_dir={"": "src"},
    setup_requires=["numpy>=1.13.3", "setuptools>=27.3", "cython>=0.29.11"],
    install_requires=[
        "numpy>=1.13.3",
        "setuptools>=27.3",
        "cython>=0.29.11",
        "astropy>=3.0",
        "pydantic>=0.28",
        "requests>=2.21.0",
        "monty==2.0.4",
        "csdmpy>=0.1.4",
    ],
    ext_modules=cythonize(ext_modules, language_level=3),
    include_package_data=True,
    zip_safe=False,
    license="BSD-3-Clause",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
    ],
)



# -*- coding: utf-8 -*-
#
#  nmr_method.pxd
#
#  Created by Deepansh J. Srivastava.
#  Contact email = deepansh2012@gmail.com
#

cdef extern from "angular_momentum.h":
    void wigner_d_matrices_from_exp_I_beta(int l, int n, void *exp_I_beta,
                                  double *wigner)


# cdef extern from "site.h":
#     ctypedef struct site:
#         unsigned int index
#         double isotropic_chemical_shift_in_Hz # Isotropic chemical shift (Hz)
#         double shielding_symmetric_anisotropy_in_Hz     # Nuclear shielding anisotropy (Hz)
#         double shielding_symmetric_asymmetry            # Nuclear shielding asymmetry
#         double shielding_symmetric_orientation[3]          # Nuclear shielding PAS to CRS euler angles (rad.)
#         double quadrupole_coupling_constant_in_Hz     # Quadrupolar coupling constant (Hz)
#         double quadrupole_asymmetry          # Quadrupolar asymmetry parameter
#         double quadrupole_orientation[3]        # Quadrupolar PAS to CRS euler angles (rad.)
#         double dipolar_coupling              # dipolar coupling sof the site


cdef extern from "isotopomer_ravel.h":
    ctypedef struct isotopomer_ravel:
        int number_of_sites                    # Number of sites
        float spin                             # The spin quantum number
        double larmor_frequency                # Larmor frequency (MHz)
        double *isotropic_chemical_shift_in_Hz # Isotropic chemical shift (Hz)
        double *shielding_anisotropy_in_Hz     # Nuclear shielding anisotropy (Hz)
        double *shielding_asymmetry            # Nuclear shielding asymmetry
        double *shielding_orientation          # Nuclear shielding PAS to CRS euler angles (rad.)
        double *quadrupole_coupling_constant_in_Hz     # Quadrupolar coupling constant (Hz)
        double *quadrupole_asymmetry          # Quadrupolar asymmetry parameter
        double *quadrupole_orientation        # Quadrupolar PAS to CRS euler angles (rad.)
        double *dipolar_couplings              # dipolar coupling stored as list of lists

cdef extern from "spinning_sidebands.h":
    void spinning_sideband_core(
        # spectrum information and related amplitude
        double * spec,
        double spectral_start,
        double spectral_increment,
        int number_of_points,

        isotopomer_ravel *ravel_isotopomer,

        int quad_second_order,                    # Quad theory for second order,
        int remove_second_order_quad_isotropic,   # remove the isotropic contribution from the
                                                  # second order quad Hamiltonian.

        # spin rate, spin angle and number spinning sidebands
        int number_of_sidebands,
        double sample_rotation_frequency_in_Hz,
        double rotor_angle_in_rad,

        # The transition as transition[0] = mi and transition[1] = mf
        double *transition,
        int geodesic_polyhedron_frequency,
        unsigned int integration_volume             # 0-octant, 1-hemisphere, 2-sphere
        )



# -*- coding: utf-8 -*-
"""Test for the base ConfigSimulator class."""
import pytest
from mrsimulator import Simulator


def test_config():

    # set config
    b = Simulator()
    error = "instance of ConfigSimulator expected"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        b.config = {}

    a = Simulator()

    assert b == a

    # number of sidebands
    assert a.config.number_of_sidebands == 64
    a.config.number_of_sidebands = 10
    assert a.config.number_of_sidebands == 10

    error = "Expecting a positive integer, found"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.number_of_sidebands = "2"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.number_of_sidebands = 1.4

    # integration density
    assert a.config.integration_density == 70
    a.config.integration_density = 20
    assert a.config.integration_density == 20

    error = "Expecting a positive integer, found"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.integration_density = -12
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.integration_density = {}

    # integration volume
    assert a.config.integration_volume == "octant"
    a.config.integration_volume = "hemisphere"
    assert a.config.integration_volume == "hemisphere"

    error = (
        "value is not a valid enumeration literal; permitted: 'octant', 'hemisphere'"
    )
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.integration_volume = "sphere"

    # decompose
    assert a.config.decompose is False
    a.config.decompose = True
    assert a.config.decompose is True

    error = "Expecting a boolean."
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        a.config.decompose = [5, 23]

    assert a.config.dict() == {
        "decompose": True,
        "number_of_sidebands": 10,
        "integration_volume": "hemisphere",
        "integration_density": 20,
    }

    assert a.config._dict == {
        "decompose": True,
        "number_of_sidebands": 10,
        "integration_volume": 1,
        "integration_density": 20,
    }

    assert str(a.config) == str(
        {
            "number_of_sidebands": 10,
            "integration_volume": "hemisphere",
            "integration_density": 20,
            "decompose": True,
        }
    )
    assert b != a



cimport nmr_methods as clib
from libcpp cimport bool as bool_t
from numpy cimport ndarray
import numpy as np
import cython

__author__ = "Deepansh J. Srivastava"
__email__ = "deepansh2012@gmail.com"


@cython.boundscheck(False)
@cython.wraparound(False)
def one_d_spectrum(dimension,
       list isotopomers,
       int verbose=0,
       int number_of_sidebands=90,
       int integration_density=72,
       bool_t decompose=False,
       int integration_volume=1):
    """

    :ivar verbose:
        The allowed values are 0, 1, and 11. When the value is 1, the output is
        printed on the screen. When the value is 11, in addition to the output
        from 1, execution time is also printed on the screen.
        The default value is 0.
    :ivar number_of_sidebands:
        The value is an integer which corresponds to the number of sidebands
        simulated in the spectrum. The default value is 90. Note, when the
        sample spinning frequency is low, computation of more sidebands may be
        required for an acceptable result. The user is advised to ensure that
        enough sidebands are requested for computation.
    :ivar integration_density:
        The value is an integer which represents the frequency of class I
        geodesic polyhedra. These polyhedra are used in calculating the
        spherical average. Presently we only use octahedral as the frequency1
        polyhedra. As the frequency of the geodesic polyhedron increases, the
        polyhedra approach a sphere geometry. For line-shape simulation, a higher
        frequency will result in a better powder averaging.
        The default value is 72.
        Read more on the `Geodesic polyhedron <https://en.wikipedia.org/wiki/Geodesic_polyhedron>`_.
    :ivar decompose:
        A boolean. If true, returns an ordered list of spectrum corresponding
        to the ordered list of isotopomers.
    """

# ---------------------------------------------------------------------
# observed spin _______________________________________________________
    dimension = dimension[0]
    isotope = dimension.isotope
    # spin quantum number of the observed spin
    cdef double spin_quantum_number = dimension.spin

    # gyromagnetic ratio
    cdef double larmor_frequency = dimension.larmor_frequency
    cdef double factor = 1.0
    if larmor_frequency < 0.0:
        factor = -1.0

    # if verbose in [1, 11]:
    #     print(f'Simulating {isotope} (I={spin_quantum_number})')
    #     print(f'Larmor frequency (ω0 = - γ B0) = {larmor_frequency/1.0e6} MHz')
    #     print((f'Recording {isotope} spectrum with {number_of_points} '
    #             f'points over {spectral_width} Hz bandwidth'))
    #     print((f"and a reference offset of {dimension['reference_offset']} Hz."))

    # transitions of the observed spin
    cdef ndarray[double, ndim=1] transition_array
    cdef int number_of_transitions
    transition_array = np.asarray([-0.5, 0.5])
    number_of_transitions = int(transition_array.size/2)
    # else:
    #     energy_level_count = int(2*spin_quantum_number+1)
    #     number_of_transitions = energy_level_count-1
    #     energy_states = np.arange(energy_level_count) - spin_quantum_number
    #     transitions = [ [energy_states[i], energy_states[i+1]] for i in range(number_of_transitions)]
    #     transition_array = np.asarray(transitions).ravel()


# Generate the dimension coordinates __________________________________________________
    cdef int number_of_points = dimension.number_of_points
    cdef double spectral_width = dimension.spectral_width
    cdef double increment = spectral_width/number_of_points
    cdef double reference_offset = dimension.reference_offset *factor

    offset = increment*int(number_of_points/2.0)
    reference_offset -= offset
    # freq = dimension.coordinates_ppm

# create MRS_plan _____________________________________________________________________
    cdef double sample_rotation_frequency_in_Hz = dimension.rotor_frequency
    cdef double rotor_angle_in_rad = dimension.rotor_angle

    # B0 = dimension.magnetic_flux_density

    # if verbose in [1, 11]:
    #     text = "`one_d_spectrum` method simulation parameters."
    #     len_ = len(text)
    #     print(text)
    #     print(f"{'-'*(len_-1)}")
    #     print (f'Macroscopic magnetic flux density (B0) = {B0} T')
    #     print (f'Sample rotation angle is (θ) = {rotor_angle_in_rad} rad')
    #     print (f'Sample rotation frequency (𝜈r) = {sample_rotation_frequency_in_Hz} Hz')

# sites _______________________________________________________________________________
    # CSA
    cdef int number_of_sites
    cdef ndarray[double] iso_n
    cdef ndarray[double] zeta_n
    cdef ndarray[double] eta_n
    cdef ndarray[double] ori_n

    # quad
    cdef ndarray[double] Cq_e
    cdef ndarray[double] eta_e
    cdef ndarray[double] ori_e

    cdef ndarray[double] D_c

    cdef int i, trans__
    cdef ndarray[double, ndim=1] amp
    amp1 = np.zeros(number_of_points, dtype=np.float64)
    amp_individual = []

    cdef clib.isotopomer_ravel isotopomer_struct
    # cdef clib.isotopomers_list *isotopomers_list_c

    # ---------------------------------------------------------------------
    # sample _______________________________________________________________
    for index_isotopomer, isotopomer in enumerate(isotopomers):
        abundance = isotopomer['abundance']
        sub_sites = [site for site in isotopomer['sites'] if site['isotope'] == isotope]

        number_of_sites= len(sub_sites)

        # site specification
        # CSA
        iso_n = np.empty(number_of_sites, dtype=np.float64)
        zeta_n = np.empty(number_of_sites, dtype=np.float64)
        eta_n = np.empty(number_of_sites, dtype=np.float64)
        ori_n = np.zeros(3*number_of_sites, dtype=np.float64)

        # quad
        Cq_e = np.zeros(number_of_sites, dtype=np.float64)
        eta_e = np.zeros(number_of_sites, dtype=np.float64)
        ori_e = np.zeros(3*number_of_sites, dtype=np.float64)

        D_c = np.zeros(number_of_sites, dtype=np.float64)

        # cdef int i, size = len(sample)

        amp = np.zeros(number_of_points*number_of_sites)

        for i in range(number_of_sites):
            site = sub_sites[i]


            # CSA tensor
            iso = site['isotropic_chemical_shift']
            if iso is None: iso = 0.0

            shielding = site['shielding_symmetric']
            if shielding is not None:
                zeta = shielding['zeta']
                if zeta is None: zeta = 0.0
                eta = shielding['eta']
                if eta is None: eta = 0.0
                alpha, beta, gamma = shielding['alpha'], shielding['beta'], shielding['gamma']
                if alpha is None: alpha = 0.0
                if beta is None: beta = 0.0
                if gamma is None: gamma = 0.0
            else:
                zeta, eta, alpha, beta, gamma = 0.0, 0.0, 0.0, 0.0, 0.0

            # if verbose in [1, 11]:
            #     text = ((
            #         f"\n{isotope} site {i} from isotopomer {index_isotopomer} "
            #         f"@ {abundance}% abundance"
            #     ))
            #     len_ = len(text)
            #     print(text)
            #     print(f"{'-'*(len_-1)}")
            #     print(f'Isotropic chemical shift (δ) = {str(1e6*iso/larmor_frequency)} ppm')
            #     print(f'Shielding anisotropy (ζ) = {str(1e6*zeta/larmor_frequency)} ppm')
            #     print(f'Shielding asymmetry (η) = {eta}')
            #     print(f'Shielding orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')

            # CSA tensor in Hz
            iso_n[i] = iso #* factor
            zeta_n[i] = zeta #* factor
            eta_n[i] = eta
            ori_n[3*i:3*i+3] = [alpha, beta, gamma]

            # quad tensor
            if spin_quantum_number > 0.5:
                quad = site['quadrupolar']
                if quad is not None:
                    Cq = quad['Cq']
                    if Cq is None: Cq = 0.0
                    eta = quad['eta']
                    if eta is None: eta = 0.0
                    alpha, beta, gamma = quad['alpha'], quad['beta'], quad['gamma']
                    if alpha is None: alpha = 0.0
                    if beta is None: beta = 0.0
                    if gamma is None: gamma = 0.0
                else:
                    Cq, eta, alpha, beta, gamma = 0.0, 0.0, 0.0, 0.0, 0.0

                Cq_e[i] = Cq #* factor
                eta_e[i] = eta
                ori_e[3*i:3*i+3] = [alpha, beta, gamma]

                # if verbose in [1, 11]:
                #     print(f'Quadrupolar coupling constant (Cq) = {Cq_e[i]/1e6} MHz')
                #     print(f'Quadrupolar asymmetry (η) = {eta}')
                #     print(f'Quadrupolar orientation = [alpha = {alpha}, beta = {beta}, gamma = {gamma}]')


        if number_of_sites != 0:
            isotopomer_struct.number_of_sites = number_of_sites
            isotopomer_struct.spin = spin_quantum_number
            isotopomer_struct.larmor_frequency = larmor_frequency

            isotopomer_struct.isotropic_chemical_shift_in_Hz = &iso_n[0]
            isotopomer_struct.shielding_anisotropy_in_Hz = &zeta_n[0]
            isotopomer_struct.shielding_asymmetry = &eta_n[0]
            isotopomer_struct.shielding_orientation = &ori_n[0]

            isotopomer_struct.quadrupole_coupling_constant_in_Hz = &Cq_e[0]
            isotopomer_struct.quadrupole_asymmetry = &eta_e[0]
            isotopomer_struct.quadrupole_orientation = &ori_e[0]

            isotopomer_struct.dipolar_couplings = &D_c[0]

            # isotopomers_list_c[i] = isotopomer_struct
            for trans__ in range(number_of_transitions):
                # spin transitions
                # m_initial = transition_array[i][0]
                # m_final   = transition_array[i][1]
                clib.spinning_sideband_core(
                    # spectrum information and related amplitude
                    &amp[0],
                    reference_offset,
                    increment,
                    number_of_points,

                    &isotopomer_struct,

                    1, # quad_second_order_c,
                    0, # turn off quad second order isotropic contribution

                    # spin rate, spin angle and number spinning sidebands
                    number_of_sidebands,
                    sample_rotation_frequency_in_Hz,
                    rotor_angle_in_rad,

                    &transition_array[2*trans__],
                    integration_density,
                    integration_volume,          # 0-octant, 1-hemisphere, 2-sphere.
                    )

            temp = amp.reshape(number_of_sites, number_of_points).sum(axis=0)*abundance

            ## reverse the spectrum if larmor_frequency is negative
            if larmor_frequency < 0:
                if number_of_points % 2 == 0:
                    temp[1:] = temp[1:][::-1]
                else:
                    temp = temp[::-1]

            if decompose:
                amp_individual.append(temp)
            else:
                amp1 += temp
        else:
            if decompose:
                amp_individual.append([])

    if decompose:
        amp1 = amp_individual


    return amp1
